from .ilcd import IlcdArchive, typeDirs, get_flow_ref, uuid_regex, dtype_from_nsmap
from ..xml_widgets import *
from ...entities import LcEntity, LcQuantity, LcUnit
from .quantity import IlcdQuantityImplementation
from .index import IlcdIndexImplementation
from antelope import comp_dir


def get_cf_value(exch, ns=None):
    try:
        v = float(find_tag(exch, 'resultingAmount', ns=ns))
    except ValueError:
        v = None
    return v


class IlcdLcia(IlcdArchive):
    """
    Slightly extends the IlcdArchive with a set of functions for loading LCIA factors and adding them as
    quantities + charaterizations
    """
    def make_interface(self, iface, privacy=None):
        if iface == 'index':
            return IlcdIndexImplementation(self)
        elif iface == 'quantity':
            return IlcdQuantityImplementation(self)
        else:
            return super(IlcdLcia, self).make_interface(iface)

    def _make_reference_unit(self, o, ns=None):
        """
        This is a bit of a hack. ILCD has distinct LciaMethod objects and FlowProperty objects.  The LCIA Method lists
        a FlowProperty as its reference quantity (like "mass C2H4 equivalents"), and then the flow property lists a
        reference unit (like "mass").  This is a problem for us because we consider the LciaMethod to BE a quantity,
        and so we want the FlowProperty to present as a unit.
        :param o:
        :param ns:
        :return:
        """
        ref_to_ref = find_tag(o, 'referenceQuantity', ns=ns)
        r_uuid = ref_to_ref.attrib['refObjectId']
        r_uri = ref_to_ref.attrib['uri']
        return self._check_or_retrieve_child(r_uuid, r_uri)

    def _create_lcia_quantity(self, o, ns):

        u = str(find_common(o, 'UUID'))
        try_q = self[u]
        if try_q is not None:
            lcia = try_q
        else:
            n = str(find_common(o, 'name'))

            c = str(find_common(o, 'generalComment'))

            m = '; '.join([str(x) for x in find_tags(o, 'methodology', ns=ns)])
            ic = '; '.join([str(x) for x in find_tags(o, 'impactCategory', ns=ns)])
            ii = '; '.join([str(x) for x in find_tags(o, 'impactIndicator', ns=ns)])

            ry = str(find_tag(o, 'referenceYear', ns=ns))
            dur = str(find_tag(o, 'duration', ns=ns))

            rq = self._make_reference_unit(o, ns=ns)
            ru = LcUnit('%s %s' % (rq.unit, rq['Name']), unit_uuid=rq.uuid)

            ext_ref = '%s/%s' % (typeDirs['LCIAMethod'], u)

            lcia = LcQuantity(ext_ref, referenceUnit=ru, Name=n, Comment=c, Method=m, Category=ic, Indicator=ii,
                              ReferenceYear=ry,
                              Duration=dur, UnitConversion=rq['UnitConversion'])

            self.add(lcia)

        return lcia

    def _load_all(self):
        super(IlcdLcia, self)._load_all()
        self.load_lcia()
        self.check_counter('quantity')

    def _fetch(self, term, dtype=None, version=None, **kwargs):
        o = super(IlcdLcia, self)._fetch(term, dtype=dtype, version=version, **kwargs)
        if isinstance(o, LcEntity) or o is None:
            return o
        if dtype is None:
            dtype = dtype_from_nsmap(o.nsmap)
        if dtype == 'LCIAMethod':
            ns = find_ns(o.nsmap, 'LCIAMethod')
            return self._create_lcia_quantity(o, ns)
        return o

    def _load_factor(self, ns, factor, lcia, load_all_flows=False):
        f_uuid, f_uri, f_dir = get_flow_ref(factor, ns=ns)
        if self[f_uuid] is None:
            if not load_all_flows:
                # don't bother loading factors for flows that don't exist
                return
        cf = float(find_tag(factor, 'meanValue', ns=ns))
        loc = str(find_tag(factor, 'location', ns=ns))
        if loc == '':
            loc = None
        flow = self._check_or_retrieve_child(f_uuid, f_uri)
        cx = self.tm[flow.context]
        if cx.sense is None:
            cx.sense = {'Input': 'Source', 'Output': 'Sink'}[f_dir]
        else:
            if comp_dir(cx.sense) != f_dir:
                print('flow %s: context %s sense %s conflicts with direction %s; negating factor' % (f_uuid, cx,
                                                                                                     cx.sense, f_dir))
                cf *= -1
        return self.tm.add_characterization(flow.link, flow.reference_entity, lcia, cf, context=flow.context, location=loc)

    def load_lcia_method(self, u, version=None, load_all_flows=False):
        """

        :param u:
        :param version:
        :param load_all_flows: [False] If False, load CFs only for already-loaded flows. If True, load all flows
        :return:
        """
        o = self._get_objectified_entity(self._path_from_parts('LCIAMethod', u, version=version))
        ns = find_ns(o.nsmap, 'LCIAMethod')

        lcia = self._create_lcia_quantity(o, ns)

        if load_all_flows is not None:
            for factor in o['characterisationFactors'].getchildren():  # British spelling! brits aren't even IN the EU anymore
                self._load_factor(ns, factor, lcia, load_all_flows=load_all_flows)
        return lcia

    def load_lcia(self, **kwargs):
        for f in self.list_objects('LCIAMethod'):
            u = uuid_regex.search(f).groups()[0]
            try:
                self._get_entity(u)
            except KeyError:
                self.load_lcia_method(u, **kwargs)

    def generate_factors(self, quantity):
        o = self._get_objectified_entity(self._path_from_ref(quantity))
        ns = find_ns(o.nsmap, 'LCIAMethod')

        lcia = self._create_lcia_quantity(o, ns)

        for factor in o['characterisationFactors'].getchildren():
            yield self._load_factor(ns, factor, lcia, load_all_flows=True)



from ...archives import BasicArchive, Qdb
from ...lcia_engine import IPCC_2007_GWP
from antelope import EntityNotFound, ConversionReferenceMismatch  # , NoFactorsFound
from ..quantity import RefQuantityRequired

import unittest

ar = BasicArchive.from_file(IPCC_2007_GWP)
gwp = ar['Global Warming Air']
mass = ar['Mass']

volu_uuid = '93a60a56-a3c8-22da-a746-0800200c9a66'


class QuantityImplementation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.I = Qdb.new()
        cls.qq_traci = ar.make_interface('quantity')
        cls.gwp = cls.qq_traci['Global Warming Air'].make_ref(cls.qq_traci)

    def test_bad_ref(self):
        with self.assertRaises(RefQuantityRequired):
            self.qq_traci.cf('hfc-134', self.gwp)

    def test_gwp_factor(self):
        self.assertEqual(self.qq_traci.cf('hfc-134', self.gwp, ref_quantity=mass, context='air'), 1100.0)
        self.assertEqual(self.qq_traci.cf('hfc-143', self.gwp, ref_quantity=mass, context='air'), 353.0)

    def test_quantity_cf_flowable(self):
        cf = list(self.gwp.factors(flowable='hfc-134'))
        self.assertEqual(len(cf), 1)
        self.assertEqual(cf[0].value, 1100.0)

    def test_list_of_flowables(self):
        cf = list(self.gwp.factors())
        self.assertEqual(len(cf), 91)

    def test_q_relation(self):
        self.assertEqual(self.gwp.cf('carbon dioxide', ref_quantity=mass, context='air'), 1.0)
        self.assertEqual(self.gwp.cf('nitrous oxide', ref_quantity=mass, context='air'), 298.0)
        self.assertEqual(self.gwp.cf('10024-97-2', ref_quantity=mass, context='air'), 298.0)
        self.assertEqual(self.gwp.quantity_relation(mass, 'carbon dioxide', 'water').value, 0.0)
        # this will only work with an LciaEngine
        ilcd_vol = self.I[volu_uuid]
        ar.add(ilcd_vol)  ## no longer done automatically in get_canonical
        with self.assertRaises(ConversionReferenceMismatch):
            self.gwp.quantity_relation(ilcd_vol, 'carbon tetrachloride', 'air')


if __name__ == '__main__':
    unittest.main()



from ..archives import Qdb, REF_QTYS
from .lcia_engine import LciaEngine, DEFAULT_CONTEXTS, DEFAULT_FLOWABLES
from antelope import QuantityRef, FlowInterface

import os

IPCC_2007_GWP = os.path.join(os.path.dirname(__file__), 'data', 'ipcc_2007_gwp.json')


class LciaDb(Qdb):
    """
    Augments the Qdb with an LciaEngine instead of a TermManager
    """
    @classmethod
    def new(cls, source=REF_QTYS, **kwargs):
        lcia = LciaEngine(**kwargs)
        qdb = cls.from_file(source, term_manager=lcia, quiet=True)
        return qdb

    def _add_char(self, flow, q, v):
        """
        For LciaDb, all factors loaded from JSON become local-- let's see if this causes any problems
        :param flow:
        :param q:
        :param v:
        :return:
        """
        self.tm.add_characterization(flow.link, flow.reference_entity, q, v, context=flow.context,
                                     origin=self.ref)

    def _ref_to_key(self, key):
        """
        LciaDb uses links as keys so as to store different-sourced versions of the same quantity. But we also want
        to find local entities by external ref- so if they come up empty we try prepending local origin.
        of course that won't work basically ever, since none of the canonical quantities have local origin.
        so this may require some tuning.
        :param key:
        :return:
        """
        key = super(LciaDb, self)._ref_to_key(key)
        if key is None:
            key = super(LciaDb, self)._ref_to_key('%s/%s' % (self.ref, key))
        return key

    def __getitem__(self, item):
        """
        Note: this user-friendliness check adds 20% to the execution time of getitem-- so avoid it if possible
        (use _get_entity directly -- especially now that upstream is now deprecated)
        (note that _get_entity does not get contexts)

        :param item:
        :return:
        """
        if hasattr(item, 'link'):
            item = item.link
        return super(LciaDb, self).__getitem__(item)

    def _ensure_valid_refs(self, entity):
        if entity.origin is None:
            entity.origin = self.ref
            # raise AttributeError('Origin not set! %s' % entity)
        super(LciaDb, self)._ensure_valid_refs(entity)

    '''
    def add(self, entity):
        """
        Add entity to archive.  If entity is a quantity ref, add a masquerade to the lcia engine
        :param entity:
        :return:
        """
        try:
            self._add(entity, entity.link)
        except EntityExists:
            # merge incoming entity's properties with existing entity
            current = self[entity.link]
            current.merge(entity)
            return

        if entity.uuid is not None:
            self._entities[entity.uuid] = entity

        self._add_to_tm(entity)
    '''

    def _add_to_tm(self, entity, merge_strategy=None):
        if entity.entity_type == 'quantity':
            if entity.is_lcia_method:
                ind = entity['Indicator']
            else:
                ind = None
            if entity.is_entity and not entity.configured:  # local db is authentic source - do not masquerade
                # print('LciaDb: Adding real entity %s' % entity.link)
                q_masq = QuantityRef(entity.external_ref, self.query, entity.reference_entity,
                                     Name=entity['Name'], Indicator=ind)
                entity.set_qi(self.make_interface('quantity'))
            else:  # ref -- masquerade
                # print('LciaDb: Adding qty ref %s' % entity)
                q_masq = QuantityRef(entity.external_ref, self.query, entity.reference_entity, masquerade=entity.origin,
                                     Name=entity['Name'], Indicator=ind)

            for k in entity.properties():  # local only for ref
                q_masq[k] = entity[k]

            # print('LciaDb: Adding masquerade %s' % q_masq)
            self.tm.add_quantity(q_masq)
            assert self.tm.get_canonical(q_masq) is q_masq, 'impostor:%s\noriginal:%s' % (self.tm.get_canonical(q_masq),
                                                                                          q_masq)
            self.tm.add_quantity(entity)  # should turn up as a child
            assert self.tm.get_canonical(entity) is q_masq, 'child:%s\n masq:%s[%s]' % (self.tm.get_canonical(entity),
                                                                                        q_masq, q_masq.link)

            if not entity.is_entity:  # not local -- local ones were already imported
                # print('LciaDb: Importing factors')
                self.tm.save_for_later(entity)

        elif isinstance(entity, FlowInterface):
            self.tm.add_flow(entity, merge_strategy=merge_strategy)

    def _serialize_quantities(self, domesticate=False):
        """
        Do not save masqueraded quantities
        :param domesticate:
        :return:
        """
        return sorted([q.serialize(domesticate=domesticate, drop_fields=self._drop_fields['quantity'])
                       for q in self.entities_by_type('quantity') if q.is_entity],
                      key=lambda x: x['externalId'])



import os
import re
from collections import defaultdict
from .entity_store import EntityStore, SourceAlreadyKnown, EntityExists
from .term_manager import TermManager

from ..implementations import BasicImplementation, IndexImplementation, QuantityImplementation, ConfigureImplementation
from antelope import BasicQuery, EntityNotFound, FlowInterface
from ..entities import LcQuantity, LcUnit, LcFlow
from ..characterizations import DuplicateCharacterizationError

from ..from_json import from_json, to_json



class OldJson(Exception):
    pass


class ContextCollision(Exception):
    pass


class InterfaceError(Exception):
    pass


class ArchiveError(Exception):
    pass


LD_CONTEXT = 'https://bkuczenski.github.io/lca-tools-datafiles/context.jsonld'

BASIC_ENTITY_TYPES = ('quantity', 'flow')


'''
LcArchive Stored Configuration.

add these objects with archive.add_config()
applied in sequence with archive.apply_config()
'''


class BasicArchive(EntityStore):
    """
    Adds on basic functionality to the archive interface: add new entities; deserialize entities.

    The Basic Archive also introduces the Term Manager, which is used to handle flow characterization information.

    Each archive must have a term manager; however, the same term manager can be shared among several archives.  If
    no TermManager is provided at instantiation as an input argument, then the archive will create a captive term
    manager that only includes entities from that archive.

    ALternatively, a term manager can be created first and supplied as an argument to each created archive, which would
    then allow them to share their terms in common with one another.

    The BasicArchive should be used for all archives that only contain flows and quantities (and contexts in the future)

    """
    _entity_types = BASIC_ENTITY_TYPES

    _drop_fields = defaultdict(list)  # dict mapping entity type to fields that should be omitted from serialization

    _skip_msg = set()

    @classmethod
    def from_file(cls, filename, ref=None, **init_args):
        """
        BasicArchive factory from minimal dictionary.  Must include at least one of 'dataSource' or 'dataReference'
        fields and 0 or more flows or quantities; but note that any flow present must have its reference
        quantities included. The method is inherited by LcArchives which permit processes as well; any process must
        have its exchanged flows (and their respective quantities) included.
        :param filename: The name of the file to be loaded
        :param ref: fallback reference to use if none is specified in source file
        :return:
        """
        j = from_json(filename)
        init_args.update(j.pop('initArgs', {}))

        old_ref = j.pop('dataReference', ref)
        ref = init_args.pop('dataReference', old_ref)
        ns_uuid = j.pop('nsUuid', j.pop('ns_uuid', None))  # this is for opening legacy files
        if ns_uuid is None:
            ns_uuid = init_args.pop('ns_uuid', None)
        ar = cls(filename, ref=ref, ns_uuid=ns_uuid, **init_args)
        ar.load_from_dict(j, jsonfile=filename)
        return ar

    @classmethod
    def from_already_open_file(cls, j, filename, ref=None, **kwargs):
        """
        This is an in-between function that should probably be refactored away / folded into archive_from_json (which
        is the only place it's used)
        :param j:
        :param filename:
        :param ref: incoming ref from catalog
        :param kwargs:
        :return:
        """
        init_args = j.pop('initArgs', {})
        ns_uuid = j.pop('nsUuid', j.pop('ns_uuid', None))  # this is for opening legacy files
        if ns_uuid is None:
            ns_uuid = init_args.pop('ns_uuid', None)
        kwargs.update(init_args)

        old_ref = j.pop('dataReference', ref)
        existing_ref = kwargs.pop('dataReference', old_ref)  # this will be the latest of init[dataRef], [dataRef], ref

        source = j.pop('dataSource')
        ar = cls(source, ref=existing_ref, ns_uuid=ns_uuid, static=True, **kwargs)
        if ref != ar.ref:
            ar.set_origin(ref)

        ar.load_from_dict(j, jsonfile=filename)

        return ar

    def __init__(self, *args, contexts=None, flowables=None, term_manager=None, **kwargs):
        super(BasicArchive, self).__init__(*args, **kwargs)
        self._tm = term_manager or TermManager(contexts=contexts, flowables=flowables)

    @property
    def query(self):
        return BasicQuery(self)

    @property
    def tm(self):
        return self._tm

    def _check_key_unused(self, key):
        """
        If the key is unused, return the UUID. Else raise EntityExists
        :param key:
        :return:
        """
        u = self._ref_to_uuid(key)
        try:
            e = self._get_entity(u)
        except KeyError:
            return u
        raise EntityExists(str(e))

    '''
    def new_quantity(self, name, ref_unit, **kwargs):
        u = self._check_key_unused(name)
        q = LcQuantity(u, ref_unit=LcUnit(ref_unit), Name=name, origin=self.ref, external_ref=name, **kwargs)
        self.add(q)
        return q

    def new_flow(self, name, ref_qty, CasNumber='', **kwargs):
        u = self._check_key_unused(name)
        f = LcFlow(u, Name=name, ReferenceQuantity=ref_qty, CasNumber=CasNumber, origin=self.ref, external_ref=name,
                   **kwargs)
        self.add_entity_and_children(f)
        return f
    '''

    def make_interface(self, iface):
        if iface == 'basic':
            return BasicImplementation(self)
        elif iface == 'quantity':
            return QuantityImplementation(self)
        elif iface == 'index':
            return IndexImplementation(self)
        elif iface == 'configure':
            return ConfigureImplementation(self)
        elif iface == 'inventory':
            return ValueError('inventory is no longer a valid interface name')
        else:
            raise InterfaceError('Unable to create interface %s' % iface)

    def _ensure_valid_refs(self, entity):
        if self.tm.is_context(entity.external_ref):
            raise ContextCollision('Entity external_ref %s is already known as a context identifier' %
                                   entity.external_ref)
        super(BasicArchive, self)._ensure_valid_refs(entity)

    def add(self, entity):
        self._add(entity, entity.external_ref)
        if hasattr(entity, 'uuid') and entity.uuid is not None:  # BasicArchives: allow UUID to retrieve entity as well, if defined
            self._entities[entity.uuid] = entity

        self._add_to_tm(entity)

    def _add_to_tm(self, entity, merge_strategy=None):
        if entity.entity_type == 'quantity':
            self.tm.add_quantity(entity)
            if entity.is_entity and not entity.configured:  # not ref
                try:
                    entity.set_qi(self.make_interface('quantity'))
                except InterfaceError:
                    pass  # quantities in index archives will not be able to use the quantity impl
        elif isinstance(entity, FlowInterface):
            # characterization infrastructure
            self.tm.add_flow(entity, merge_strategy=merge_strategy)

    def __getitem__(self, item):
        """
        Note: this user-friendliness check adds 20% to the execution time of getitem-- so avoid it if possible
        (use _get_entity directly -- especially now that upstream is now deprecated)
        (note that _get_entity does not get contexts)

        :param item:
        :return:
        """
        # cx = self.tm.__getitem__(item)
        # if cx is None:
        if hasattr(item, 'external_ref'):
            item = item.external_ref
        return super(BasicArchive, self).__getitem__(item)
        # return cx

    def _add_children(self, entity):
        if entity.entity_type == 'quantity' and entity.is_entity:
            # reset unit strings- units are such a hack
            if isinstance(entity.reference_entity, LcUnit):
                entity.reference_entity._external_ref = entity.reference_entity.unitstring
        elif entity.entity_type == 'flow':
            self.add_entity_and_children(entity.reference_entity)

    def add_entity_and_children(self, entity):
        try:
            self.add(entity)
        except EntityExists:
            return
        self._add_children(entity)

    def _create_unit(self, unitstring):
        """
        This returns two things: an LcUnit having the given unit string, and a dict of conversion factors
        (or None if the class doesn't support it).  The dict should have unit strings as keys, and the values should
        have the property that each key-value pair has the same real magnitude.  In other words, the [numeric] values
        should report the number of [keys] that is equal to the reference unit.  e.g. for a reference unit of 'kg',
        the UnitConversion entry for 'lb' should have the value 2.2046... because 2.2046 lb = 1 kg

        In many cases, this will require the supplied conversion value to be inverted.

        The conversion dict should be stored in the Quantity's UnitConversion property.  See IlcdArchive for an
        example implementation.
        :param unitstring:
        :return:
        """
        return LcUnit(unitstring), None

    def _quantity_from_json(self, entity_j, ext_ref):
        # can't move this to entity because we need _create_unit- so we wouldn't gain anything
        unit, _ = self._create_unit(entity_j.pop('referenceUnit'))
        entity_j['referenceUnit'] = unit
        quantity = LcQuantity(ext_ref, **entity_j)
        return quantity

    def _flow_from_json(self, entity_j, ext_ref):
        chars = entity_j.pop('characterizations', [])
        if 'referenceQuantity' in entity_j:
            rq = entity_j.pop('referenceQuantity')
        else:
            rq = next(c['quantity'] for c in chars if 'isReference' in c and c['isReference'] is True)
        if rq is None:
            print('Warning: no reference quantity for flow %s' % ext_ref)
            ref_q = None
        else:
            try:
                ref_q = self.tm.get_canonical(rq)
            except EntityNotFound:
                ref_q = self._get_entity(rq)
        return LcFlow(ext_ref, referenceQuantity=ref_q, **entity_j)

    def _add_char(self, flow, q, v):
        self.tm.add_characterization(flow.link, flow.reference_entity, q, v, context=flow.context,
                                     origin=flow.origin)

    def _add_chars(self, flow, chars):
        for c in chars:
            if 'isReference' in c:
                if c['isReference'] is True:
                    continue
            v = None
            try:
                q = self.tm.get_canonical(c['quantity'])  # this is required because of foreground; _process_from_json unaffected
            except EntityNotFound:
                if c['quantity'] not in self._skip_msg:
                    print('skipping unrecognized quantity %s' % c['quantity'])
                    self._skip_msg.add(c['quantity'])
                continue
            if q is None:
                continue
                # import json
                # import sys
                # print(ext_ref)
                # json.dump(c, sys.stdout, indent=2)
                # raise KeyError
            if 'value' in c:
                v = c['value']
            try:
                self._add_char(flow, q, v)
            except DuplicateCharacterizationError:
                print('Skipping duplicate CF %3.6g %s %s' % (v, q, flow))

    def _make_entity(self, e, etype, ext_ref):
        if etype == 'quantity':
            entity = self._quantity_from_json(e, ext_ref)
        elif etype == 'flow':
            entity = self._flow_from_json(e, ext_ref)
        else:
            raise TypeError('Unknown entity type %s' % etype)
        return entity

    def entity_from_json(self, e):
        """
        Create an LcEntity subclass from a json-derived dict

        this could use some serious refactoring
        :param e:
        :return:
        """
        if 'tags' in e:
            raise OldJson('This file type is no longer supported.')
        e['entity_uuid'] = e.pop('entityId', None)
        ext_ref = e.pop('externalId', None)
        if ext_ref is None:
            if e['entity_uuid'] is None:
                raise OldJson('Missing both externalId and entityId')
            ext_ref = e['entity_uuid']
        ext_ref = str(ext_ref)
        if ext_ref in self._entities:
            return self[ext_ref]
        etype = e.pop('entityType')
        if etype == 'flow':
            # need to delay adding characterizations until after entity is registered with term manager
            try:
                chars = e['characterizations']
            except KeyError:
                chars = []
        else:
            chars = []
        e['origin'] = e.pop('origin', self.ref)

        entity = self._make_entity(e, etype, ext_ref)

        self.add(entity)
        if etype == 'flow':
            # characterization infrastructure
            self._add_chars(entity, chars)

        return entity

    def load_from_dict(self, j, _check=True, jsonfile=None):
        """
        Archives loaded from JSON files are considered static.
        :param j:
        :param _check: whether to run check_counter to print out statistics at the end
        :param jsonfile: [None] if present, add to the list of sources for the canonical ref
        :return:
        """

        if 'catalogNames' in j:
            for ref, l in j['catalogNames'].items():
                if isinstance(l, str) or l is None:
                    self._add_name(ref, l)
                else:
                    for s in l:
                        self._add_name(ref, s)

        if 'dataReference' in j:
            # new style
            source = j['dataSource']
            ref = j['dataReference']
            try:
                self._add_name(ref, source)
            except SourceAlreadyKnown:
                self._add_name(ref, None)

        if jsonfile is not None:
            self._add_name(self.ref, jsonfile, rewrite=True)

        q_map = dict()
        if 'quantities' in j:
            for e in j['quantities']:
                e['entityType'] = 'quantity'
                q = self.entity_from_json(e)
                if q.uuid is not None:
                    q_map[q.uuid] = q
                q_map[q.external_ref] = q
        if 'termManager' in j:
            self.tm.add_from_json(j['termManager'], q_map, self.ref)

        if 'flows' in j:
            for e in j['flows']:
                e['entityType'] = 'flow'
                self.entity_from_json(e)

        if 'loaded' in j:
            self._loaded = j['loaded']

        if _check:
            self.check_counter()

    def _load_all(self, **kwargs):
        if self.source is None:
            return
        if os.path.exists(self.source):
            self.load_from_dict(from_json(self.source))

    @staticmethod
    def _narrow_search(entity, **kwargs):
        """
        Narrows a result set using sequential keyword filtering
        :param entity:
        :param kwargs:
        :return: bool
        """
        def _recurse_expand_subtag(tag):
            if tag is None:
                return ''
            elif isinstance(tag, str):
                return tag
            else:
                return '\n'.join([_recurse_expand_subtag(t) for t in tag])
        keep = True
        for k, v in kwargs.items():
            if v is None:
                continue
            if not entity.has_property(k):
                return False
            if isinstance(v, str):
                v = [v]
            for vv in v:
                keep = keep and bool(re.search(vv, _recurse_expand_subtag(entity[k]),
                                               flags=(re.IGNORECASE|re.MULTILINE)))
        return keep

    def search(self, etype=None, upstream=False, **kwargs):
        """
        Find entities by search term, either full or partial uuid or entity property like 'Name', 'CasNumber',
        or so on.
        :param etype: optional first argument is entity type
        :param upstream: (False) if upstream archive exists, search there too
        :param kwargs: regex search through entities' properties as named in the kw arguments
        :return: result set
        """
        if etype is None:
            if 'entity_type' in kwargs.keys():
                etype = kwargs.pop('entity_type')
        if etype is not None:
            for ent in self.entities_by_type(etype):
                if self._narrow_search(ent, **kwargs):
                    yield ent
        else:
            for ent in self._entities.values():
                if self._narrow_search(ent, **kwargs):
                    yield ent
        if upstream and self._upstream is not None:
            self._upstream.search(etype, upstream=upstream, **kwargs)

    def _serialize_quantities(self, domesticate=False):
        return sorted([q.serialize(domesticate=domesticate, drop_fields=self._drop_fields['quantity'])
                       for q in self.entities_by_type('quantity')],
                      key=lambda x: x['externalId'])

    def serialize(self, characterizations=False, values=False, domesticate=False):
        """

        :param characterizations:
        :param values:
        :param domesticate: [False] if True, omit entities' origins so that they will appear to be from the new archive
         upon serialization
        :return:
        """
        j = super(BasicArchive, self).serialize()
        j['@context'] = LD_CONTEXT

        j['flows'] = sorted([f.serialize(domesticate=domesticate, drop_fields=self._drop_fields['flow'])
                             for f in self.entities_by_type('flow')],
                            key=lambda x: x['externalId'])
        if characterizations:
            j['termManager'], qqs, rqs = self.tm.serialize(self.ref, values=values)
            for q in qqs:
                if self[q] is None:
                    self.add(self.tm.get_canonical(q))

        j['quantities'] = self._serialize_quantities(domesticate=domesticate)

        return j

    def export_quantities(self, filename, *quantities, domesticate=True, values=True, gzip=False):
        """

        :param filename:
        :param quantities: positional args are quantities to serialize
        :param domesticate: [True- new archive becomes reference]  if False, the qty source origin will remain canonical
        :param values: [True- new archive stores cfs] if False the export will only include flowable+context lists
        :param gzip: [False] whether to gzip the file
        :return:
        """
        j = super(BasicArchive, self).serialize()
        j['dataSourceType'] = 'BasicArchive'
        j['dataSource'] = filename
        j['termManager'], qq, rq = self.tm.serialize(self.ref, *quantities, values=values)
        qs = [self[u] for u in rq]
        for ref in qq:
            q = self[ref]
            if q not in qs:
                qs.append(q)

        j['quantities'] = sorted([q.serialize(domesticate=domesticate)
                                  for q in qs],
                                 key=lambda x: x['externalId'])
        to_json(j, filename, gzip=gzip)

    def _serialize_all(self, **kwargs):
        return self.serialize(characterizations=True, values=True, **kwargs)




from .entity_store import EntityStore, SourceAlreadyKnown, EntityExists, uuid_regex
from .basic_archive import BasicArchive, BASIC_ENTITY_TYPES, InterfaceError, ArchiveError
from .archive_index import index_archive, BasicIndex, LcIndex
from .term_manager import TermManager
from .lc_archive import LcArchive, LC_ENTITY_TYPES
from ..from_json import from_json
from ..entities.processes import NoExchangeFound

from pathlib import Path
from collections import defaultdict

# import pkgutil

REF_QTYS = str(Path(__file__).parent / 'data' / 'elcd_reference_quantities.json')


class Qdb(BasicArchive):
    """
    A simple archive that just contains the 25-odd reference (non-LCIA) quantities of the ELCD database circa v3.2
    """
    @classmethod
    def new(cls, ref='local.qdb'):
        """
        Create a Quantity database containing the ILCD reference quantities.  Specify a ref if desired.
        :param ref: ['local.qdb']
        """
        return cls.from_file(REF_QTYS, ref=ref)

    def _fetch(self, entity, **kwargs):
        return self.__getitem__(entity)

    def _load_all(self, **kwargs):
        self.load_from_dict(from_json(self.source))


def update_archive(archive, json_file):
    archive.load_from_dict(from_json(json_file), jsonfile=json_file)


# find antelope providers
init_map = {
    'basicarchive': BasicArchive,
    'basicindex': BasicIndex,
    'lcarchive': LcArchive,
    'lcindex': LcIndex
}

def archive_factory(ds_type):
    """
    Returns an archive class
    :param ds_type:
    :return:
    """
    dsl = ds_type.lower()
    if dsl in init_map:
        return init_map[dsl]
    raise ArchiveError('No provider found for %s' % ds_type)


def archive_from_json(fname, factory=archive_factory, catalog=None, **archive_kwargs):
    """
    :param fname: JSON filename
    :param factory: function returning a class
    :param catalog: [None] necessary to retrieve upstream archives, if specified
    :return: an ArchiveInterface
    """
    j = from_json(fname)

    if 'upstreamReference' in j or catalog is not None:
        print('**Upstream reference encountered: %s' % j['upstreamReference'])
        print('**XX Upstream is gone; catalog argument is deprecated\n')
    cls = factory(j.pop('dataSourceType', 'LcArchive'))
    return cls.from_already_open_file(j, fname, quiet=True, **archive_kwargs)


def create_archive(source, ds_type, factory=archive_factory, **kwargs):
    """
    Create an archive from a source and type specification.
    :param source:
    :param ds_type:
    :param factory: override archive factory with fancier version
    :param kwargs:
    :return:
    """
    if ds_type.lower() == 'json':
        a = archive_from_json(source, factory=factory, **kwargs)
    else:
        cls = factory(ds_type)
        a = cls(source, **kwargs)
    return a


class CheckTerms(object):
    def __init__(self, query):
        self._check = defaultdict(list)
        self._p = 0
        self._rx = 0
        self._x = 0

        for p in query.processes():
            self._p += 1
            for rx in p.references():
                self._rx += 1
                for x in p.inventory(rx):
                    self._x += 1
                    if x.type == 'node':
                        try:
                            query.get(x.termination).reference(x.flow)
                            self._check['terminated'].append(x)
                        except NoExchangeFound:
                            self._check['missing'].append(x)
                    elif x.type == 'elementary':
                        self._check['elementary'].append(x)
                    elif x.type == 'context':
                        tg = list(query.targets(x.flow))
                        if len(tg) == 0:
                            self._check['cutoff'].append(x)
                        elif len(tg) > 1:
                            self._check['ambiguous'].append(x)
                        else:
                            self._check['terminated'].append(x)
                    else:
                        self._check[x.type].append(x)
        self.show()

    def show(self):
        print('%d processes\n%d reference exchanges\n%d dependent exchanges' % (self._p, self._rx, self._x))
        for k, v in self._check.items():
            if len(v) > 0:
                print('%s: %d exchanges' % (k, len(v)))

    def exchanges(self, key):
        return self._check[key]



from __future__ import print_function, unicode_literals
import uuid

from synonym_dict import LowerDict
from antelope import QuantityRef, convert, EntityNotFound

from .entities import LcEntity


class QuantityAlreadyConfigured(Exception):
    pass


def new_quantity(name, ref_unit, external_ref=None, **kwargs):
    entity_uuid = kwargs.pop('entity_uuid', None)
    if entity_uuid is None:
        entity_uuid = kwargs.pop('uuid', None)
    if external_ref is None:
        if entity_uuid is None:
            return LcQuantity.new(name, ref_unit, **kwargs)
    return LcQuantity(external_ref, Name=name, ReferenceUnit=LcUnit(ref_unit), entity_uuid=entity_uuid, **kwargs)


class LcQuantity(LcEntity):

    _ref_field = 'referenceUnit'
    _new_fields = ()

    @classmethod
    def new(cls, name, ref_unit, **kwargs):
        """
        :param name: the name of the quantity
        :param ref_unit: the string representation of the reference unit for the quantity
        :return:
        """
        u = uuid.uuid4()
        return cls(str(u), entity_uuid=u, Name=name, ReferenceUnit=LcUnit(ref_unit), **kwargs)

    def __init__(self, external_ref, **kwargs):
        self._is_lcia = False
        super(LcQuantity, self).__init__('quantity', external_ref, **kwargs)
        if not self.has_property('UnitConversion'):
            self._d['UnitConversion'] = {self.unit: 1.0}
        self._qi = None

    def __setitem__(self, key, value):
        if key.lower() == 'indicator' and len(value) > 0:
            self._new_fields = ('Indicator', )
            self._is_lcia = True
        super(LcQuantity, self).__setitem__(key, value)

    def _set_reference(self, ref_entity):
        if isinstance(ref_entity, str):
            ref_entity = LcUnit(ref_entity)
        super(LcQuantity, self)._set_reference(ref_entity)

    def quantity_terms(self):
        yield self['Name']
        yield self.name
        yield str(self)  # this is the same as above for entities, but includes origin for refs
        yield self.external_ref  # do we definitely want this?  will squash versions together
        if self.uuid is not None:
            yield self.uuid
        if self.origin is not None:
            yield self.link
        if self.has_property('Synonyms'):
            syns = self['Synonyms']
            if isinstance(syns, str):
                yield syns
            else:
                for syn in syns:
                    yield syn

    @property
    def configured(self):
        return self._qi is not None

    def set_qi(self, qi):
        """
        Configures the quantity to access its native term manager.  Can only be set once; otherwise ignored.
        :param qi:
        :return:
        """
        if self._qi is None:
            self._qi = qi
        else:
            raise QuantityAlreadyConfigured(self)

    @property
    def unit(self):
        return self.reference_entity.unitstring

    @property
    def is_lcia_method(self):
        return self._is_lcia

    def add_synonym(self, k):
        if self.has_property('Synonyms'):
            if isinstance(self['Synonyms'], str):
                syns = {self['Synonyms']}
            else:
                syns = set(self['Synonyms'])
        else:
            syns = set()
        syns.add(k)
        self['Synonyms'] = syns


    """
    Quantity Interface Methods
    Quantity entities use the quantity interface provided by the parent archive; this emulates the operation of 
    quantity refs, which have access to the catalog.
    """
    def cf(self, flow, locale='GLO', **kwargs):
        """
        The semantics here may be confusing, but cf is flow-centered. It converts reports the amount in self that
        corresponds to a unit of the flow's reference quantity.
        :param flow:
        :param locale:
        :param kwargs:
        :return:
        """
        return self._qi.cf(flow, self, locale=locale, **kwargs)

    def characterize(self, flowable, ref_quantity, value, **kwargs):
        return self._qi.characterize(flowable, ref_quantity, self, value, **kwargs)

    def factors(self, flowable=None, context=None, **kwargs):
        return self._qi.factors(self, flowable=flowable, context=context, **kwargs)

    def profile(self, flow, **kwargs):
        """
        This is a ridiculous hack because the profile doesn't depend on self at all.  So let's make it non-ridiculous
        by defaulting to self as ref_quantity, so it's not TOTALLY ridiculous
        :param flow:
        :param kwargs:
        :return:
        """
        ref_quantity = kwargs.pop('ref_quantity', self)
        return self._qi.profile(flow, ref_quantity=ref_quantity, **kwargs)

    def do_lcia(self, inventory, **kwargs):
        return self._qi.do_lcia(self, inventory, **kwargs)

    def convert(self, from_unit=None, to=None):
        return convert(self, from_unit, to)

    """
    Interior utility functions
    These are not exactly exposed by the quantity interface and maybe should be retired
    """

    def _print_ref_field(self):
        return self.reference_entity.unitstring

    def reset_unitstring(self, ustring):
        self.reference_entity.reset_unitstring(ustring)

    @property
    def _name(self):
        n = '%s [%s]' % (self._d['Name'].replace('\n', '|'), self.reference_entity.unitstring)
        if self.is_lcia_method:
            return '%s [LCIA]' % n
        return n

    @property
    def name(self):
        return self._name

    def __str__(self):
        return self._name

    def __eq__(self, other):
        if other is self:
            return True
        if isinstance(other, QuantityRef):
            try:
                return other.is_canonical(self)
            except EntityNotFound:
                pass  # other TM doesn't know us: fall through
        return super(LcQuantity, self).__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        """
        This needs to be explicit, even though it is identical to the parent class, because otherwise the type is
        considered unhashable
        :return:
        """
        if self._origin is None:
            raise AttributeError('Origin not set!')
        return hash(self.link)


class LcUnit(object):
    """
    Dummy class to store a reference to a unit definition
    Design decision: even though ILCD unitgroups have assigned UUIDs, we are not maintaining unitgroups
    """
    entity_type = 'unit'

    def __init__(self, unitstring, unit_uuid=None):
        if unit_uuid is not None:
            self._uuid = uuid.UUID(unit_uuid)
        else:
            self._uuid = None
        if isinstance(unitstring, LcUnit):
            unitstring = unitstring.unitstring
        self._unitstring = unitstring
        self._external_ref = None

    def set_external_ref(self, external_ref):
        self._external_ref = external_ref

    def get_external_ref(self):
        return '%s' % self._unitstring if self._external_ref is None else self._external_ref

    @property
    def unitstring(self):
        return self._unitstring

    def __str__(self):
        return '[%s]' % self._unitstring

    def reset_unitstring(self, ustring):
        self._external_ref = ustring
        self._unitstring = ustring



"""
This object replaces the LciaResult types spelled out in Antelope-- instead, it serializes to an LCIA result directly.

"""
from antelope import comp_dir
from .exchanges import ExchangeValue, DissipationExchange
from .autorange import AutoRange
from numbers import Number
from math import isclose
from collections import defaultdict
# from lcatools.interfaces import to_uuid


class InconsistentQuantity(Exception):
    pass


class InconsistentScenario(Exception):
    pass


class DuplicateResult(Exception):
    pass


class InconsistentScores(Exception):
    pass


class InconsistentSummaries(Exception):
    pass


def number(val):
    try:
        return '%10.3g' % val
    except TypeError:
        return '%10.10s' % '----'


class DetailedLciaResult(object):
    """
    Contains exchange, factor, result
    """
    def __init__(self, lc_result, exchange, qrresult):
        """

        :param lc_result:
        :param exchange:
        :param qrresult: meets the QRResult spec: has properties 'flowable', 'ref', 'query', 'context', 'locale',
        'origin', 'value'
        """
        if exchange.flow.unit != qrresult.ref.unit and qrresult.value != 0.0:
            print('%s: Inconsistent qty\nexch: %s\nqrr:  %s' % (self.__class__.__name__, exchange.flow.reference_entity, qrresult))
            #  raise InconsistentQuantity('%s\n%s' % (exchange.flow.reference_entity, qrresult))
        self._exchange = exchange
        self._qr = qrresult
        self._lc = lc_result

    @property
    def exchange(self):
        return self._exchange

    @property
    def factor(self):
        return self._qr

    @property
    def _dirn_adjust(self):
        if self._qr.context.sense is None:
            return 1.0
        elif comp_dir(self._qr.context.sense) == self._exchange.direction:
            return 1.0
        return -1.0

    @property
    def is_null(self):
        return self.result == 0

    @property
    def flow(self):
        return self._exchange.flow

    @property
    def name(self):
        return '; '.join(filter(None, (self.flowable, self.context)))

    @property
    def direction(self):
        return self._exchange.direction

    @property
    def flowable(self):
        return str(self._qr.flowable)

    @property
    def context(self):
        if self.exchange.termination is not None:
            return str(self.exchange.termination)
        return str(self._qr.context)

    @property
    def content(self):
        if isinstance(self._exchange, DissipationExchange):
            return self._exchange.content()
        return None

    @property
    def value(self):
        if self._exchange.value is None:
            return 0.0
        return self._exchange.value * self._lc.scale

    @property
    def result(self):
        if self._qr.value is None:
            return 0.0
        return self.value * self._dirn_adjust * self._qr.value

    def __hash__(self):
        return hash((self._exchange.process.external_ref, self._exchange.direction, self._qr.flowable, self._qr.context))

    def __eq__(self, other):
        if not isinstance(other, DetailedLciaResult):
            return False
        return (self.exchange.process.external_ref == other.exchange.process.external_ref and
                self.flowable == other.flowable and
                self.direction[0] == other.direction[0] and
                self.context == other.context)

    def __str__(self):
        if self._dirn_adjust == -1:
            dirn_mod = '*'
        else:
            dirn_mod = ' '
        return '%s%s = %-s  x %-s [%s] %s, %s' % (dirn_mod,
                                                  number(self.result * self._lc.autorange),
                                                  number(self._qr.value * self._lc.autorange),
                                                  number(self.value),
                                                  self._qr.locale,
                                                  self.flowable,
                                                  self.context)

    def serialize(self, detailed=False):
        flowname = self.flow.name
        if self.exchange.termination is not None:
            flowname = ', '.join([flowname, self.exchange.termination.name])
        return {
            'flow': self.flowable,
            'context': self.context,
            'exchange': self.value,
            'factor': self._qr.value,
            'result': self.result,
            'locale': self._qr.locale
        }


class SummaryLciaResult(object):
    """
    like a DetailedLciaResult except omitting the exchange and factor information.  This makes them totally static.
    """
    def __init__(self, lc_result, entity, node_weight, unit_score):
        """
        :param lc_result: who "owns" you. scale report by their scale.
        entity_id must either have get_uuid() or be hashable
        :param entity: a hashable identifier
        :param node_weight: stand-in for exchange value
        :param unit_score: stand-in for factor value
        """
        self.entity = entity
        self._node_weight = node_weight
        if isinstance(unit_score, Number):
            self._static_value = unit_score
            self._internal_result = None
        else:
            self._static_value = None
            self._internal_result = unit_score

        self._lc = lc_result

    def update_parent(self, lc):
        self._lc = lc

    @property
    def name(self):
        if isinstance(self.entity, tuple):
            return '; '.join(str(k) for k in self.entity)
        try:
            return self.entity.fragment.name
        except AttributeError:
            try:
                return self.entity.name
            except AttributeError:
                return str(self.entity)

    @property
    def id(self):
        try:
            return self.entity.fragment.external_ref
        except AttributeError:
            try:
                return self.entity.external_ref
            except AttributeError:
                return None


    @property
    def static(self):
        return self._static_value is not None

    @property
    def is_null(self):
        if self.static:
            if self._static_value == 0:
                return True
            return False
        return self._internal_result.is_null

    @property
    def node_weight(self):
        return self._node_weight * self._lc.scale

    @property
    def unit_score(self):
        if self.static:
            return self._static_value
        else:
            return self._internal_result.total()

    @property
    def cumulative_result(self):
        return self.node_weight * self.unit_score

    def components(self):
        if self.static:
            yield self
        else:
            for x in self._internal_result.components():
                yield x

    def __hash__(self):
        return hash(self.entity)

    def __eq__(self, other):
        if not isinstance(other, SummaryLciaResult):
            return False
        return self.entity == other.entity

    def __str__(self):
        return '%s = %-s x %-s | %s' % (number(self.cumulative_result * self._lc.autorange), number(self.node_weight),
                                        number(self.unit_score * self._lc.autorange),
                                        self.entity)

    def show(self):
        if self.static:
            print('%s' % self)
        else:
            self._internal_result.show()

    def details(self):
        if self.static:
            yield self
        else:
            for k in self._internal_result.keys():
                for d in self._internal_result[k].details():
                    yield d

    def show_detailed_result(self):
        if self.static:
            self.show()
        else:
            self._internal_result.show_components()

    def flatten(self):
        if self.static:
            return self
        return self._internal_result.flatten(_apply_scale=self.node_weight)

    def __add__(self, other):
        """
        Add two summary LCIA results together.  This only works under the following circumstances:
         * different instances of the same entity are being added (e.g. two instances of the same flow).
           In this case, the two summaries' entities must compare as equal and their unit scores must be equal.
           The node weights are added.  Scale is ignored (scale is inherited from the primary summary)

         * Two static-valued summaries are added together.  In this case, either the scores must be equal (in which case
           the node weights are summed) or the node weights must be equal, and the unit scores are summed.

        This is the sort of garbage that should be unittested.

        :param other:
        :return:
        """
        if not isinstance(other, SummaryLciaResult):
            raise TypeError('Can only add SummaryLciaResults together')
        if self._lc is not other._lc:
            raise InconsistentSummaries('These summaries do not belong to the same LciaResult')
        if self.static:
            if other.static:
                # either the node weights or the unit scores must be equal
                if self.unit_score == other.unit_score:  # make node weights add preferentially
                    unit_score = self.unit_score
                    _node_weight = self._node_weight + (other.node_weight / self._lc.scale)
                elif self.node_weight == other.node_weight:
                    _node_weight = self._node_weight
                    unit_score = self._static_value + other.unit_score
                else:
                    # in conflicts, prefer unit node weights if they exist (i.e. in aggregations)
                    if self.node_weight == 1.0 or other.node_weight == 1.0:
                        _node_weight = 1.0
                        unit_score = self.cumulative_result + other.cumulative_result
                    else:
                        # if these are richly detailed, it is wrong to add them
                        raise InconsistentScores('These summaries do not add together:\n%s\n%s' % (self, other))
            else:
                if self.unit_score == other.unit_score:
                    unit_score = other._internal_result
                    _node_weight = self._node_weight + (other.node_weight / self._lc.scale)
                else:
                    raise InconsistentScores('These summaries have different unit scores')
        elif self.unit_score == other.unit_score:
            unit_score = self._internal_result
            if other.static:
                _node_weight = self._node_weight + (other.node_weight / self._lc.scale)
            elif self._internal_result is other._internal_result:
                # this only works because terminations cache unit scores
                # just sum the node weights, ignoring our local scaling factor (DWR!)
                if self.entity == other.entity:
                    # WARNING: FragmentFlow equality does not include magnitude or node weight
                    _node_weight = self._node_weight + (other.node_weight / self._lc.scale)
                else:
                    print("entities do not match\n self: %s\nother: %s" % (self.entity, other.entity))
                    raise InconsistentSummaries
            else:
                """
                This situation is cropping up in the CalRecycle model but it appears to be kosher. I propose the 
                following test: if the two summaries are both nonstatic and (a) have the same set of components and (b) 
                have the same unit scores, then treat them as the same.
                """
                if set(k.entity for k in self.components()) == set(k.entity for k in other.components()):
                    _node_weight = self._node_weight + (other.node_weight / self._lc.scale)
                else:
                    raise InconsistentSummaries('Components differ between non-static summaries')
        else:
            print('\n%s' % self)
            print(other)
            raise InconsistentSummaries('At least one not static, and unit scores do not match')
        return SummaryLciaResult(self._lc, self.entity, _node_weight, unit_score)

    def serialize(self, detailed=False):
        j = {
            'node_weight': self.node_weight,
            'unit_score': self.unit_score,
            'result': self.cumulative_result,
            'component': self.name,
        }
        if self.id is not None:
            j['entity_id'] = self.id

        if detailed:
            if not self.static:
                f = self.flatten()  # will yield a list of aggregate lcia scores with one component each
                j['details'] = [d.serialize() for p in f.components() for d in p.details()]
        return j


class AggregateLciaScore(object):
    """
    contains an entityId which should be either a process or a fragment (fragment stages show up as fragments??)
    The Aggregate score is constructed either from individual LCIA Details (exchange value x characterization factor)
    or from summary results
    """
    static = True

    def __init__(self, lc_result, entity):
        self.entity = entity
        self._lc = lc_result
        self.LciaDetails = []  # what exactly was having unique membership protecting us from??

    def update_parent(self, lc_result):
        self._lc = lc_result

    @property
    def name(self):
        if isinstance(self.entity, tuple):
            return '; '.join(str(k) for k in self.entity)
        try:
            return self.entity.fragment.name
        except AttributeError:
            try:
                return self.entity.name
            except AttributeError:
                return str(self.entity)

    @property
    def cumulative_result(self):
        if len(self.LciaDetails) == 0:
            return 0.0
        return sum([i.result for i in self.LciaDetails])

    @property
    def is_null(self):
        for i in self.LciaDetails:
            if not i.is_null:
                return False
        return True

    '''
    def _augment_entity_contents(self, other):
        """
        when duplicate fragmentflows are aggregated, their node weights and magnitudes should be added together
        :return:
        """
        self.entity = self.entity + other
    '''

    def add_detailed_result(self, exchange, qrresult):
        d = DetailedLciaResult(self._lc, exchange, qrresult)
        '''
        if d in self.LciaDetails:  # process.uuid, direction, flow.uuid are the same
            if factor[location] != 0:
                other = next(k for k in self.LciaDetails if k == d)
                raise DuplicateResult('exchange: %s\n  factor: %s\nlocation: %s\nconflicts with %s' %
                                      (exchange, factor, location, other.factor))
            else:
                # do nothing
                return
        '''
        self.LciaDetails.append(d)

    def show(self, **kwargs):
        self.show_detailed_result(**kwargs)

    def details(self):
        """
        generator of nonzero detailed results
        :return:
        """
        for d in self.LciaDetails:
            if d.result != 0:
                yield d

    def show_detailed_result(self, key=lambda x: x.result, show_all=False):
        for d in sorted(self.LciaDetails, key=key, reverse=True):
            if d.result != 0 or show_all:
                print('%s' % d)
        # print('=' * 60)
        # print('             Total score: %g ' % self.cumulative_result)

    def __str__(self):
        return '%s  %s' % (number(self.cumulative_result * self._lc.autorange), self.entity)

    def serialize(self, detailed=False):
        j = {
            'result': self.cumulative_result,
            'component': self.name
        }
        if hasattr(self.entity, 'external_ref'):
            j['entity_id'] = self.entity.external_ref

        if detailed:
            if self.static:
                j['details'] = []
            j['details'] = [p.serialize(detailed=False) for p in self.LciaDetails]
        return j



def show_lcia(lcia_results):
    """
    Takes in a dict of uuids to lcia results, and summarizes them in a neat table
    :param lcia_results:
    :return:
    """
    print('LCIA Results\n%s' % ('-' * 60))
    for r in lcia_results.values():
        print('%10.5g %s' % (r.total(), r.quantity))


class LciaResult(object):
    """
    An LCIA result object contains a collection of LCIA results for a related set of entities, called components.  Each
     component is an AggregateLciaScore, which itself is a collection of either detailed LCIA results or summary scores.

    Each component which is a FragmentFlow represents a specific traversal scenario and is thus static.

    Each component which is a process will contain actual exchanges and factors, which are scenario-sensitive, and
     so is (theoretically) dynamic. This is not yet useful in practice.  LCIA Results are in sharp need of testing /
     refactoring.
    """
    def __init__(self, quantity, scenario=None, private=False, scale=1.0):
        """
        If private, the LciaResult will not return any unaggregated results
        :param quantity:
        :param scenario:
        :param private:
        """
        self.quantity = quantity
        self.scenario = scenario
        self._scale = scale
        self._LciaScores = dict()
        self._cutoffs = []
        self._errors = []
        self._zeros = []

        self._private = private
        self._autorange = None
        self._failed = []

    @property
    def is_null(self):
        for i in self._LciaScores.values():
            if not i.is_null:
                return False
        return True

    def set_autorange(self, value=True):
        """
        Update the AutoRange object. Should be done before results are presented, if auto-ranging is in use.

        Auto-ranging affects the following outputs:
         * any show() or printed string
         * the results of contrib_new()
        No other outputs are affected.
        :param value:
        :return:
        """
        assert isinstance(value, bool), 'cannot set autorange to %s of type %s' % (value, type(value))
        if value:
            self._autorange = AutoRange(self.range())
        else:
            self._autorange = None

    def unset_autorange(self):
        self.set_autorange(False)

    @property
    def autorange(self):
        if self._autorange is None:
            return 1.0
        else:
            return self._autorange.scale

    @property
    def unit(self):
        if self._autorange is not None:
            return self._autorange.adj_unit(self.quantity.unit)
        else:
            return self.quantity.unit

    @property
    def scale(self):
        return self._scale

    def scale_result(self, scale):
        """
        why is this a function?
        """
        self._scale *= scale

    def _match_key(self, item):
        """
        whaaaaaaaaaat is going on here
        :param item:
        :return:
        """
        if item in self._LciaScores:
            yield self._LciaScores[item]
        else:
            for k, v in self._LciaScores.items():
                if item == v.entity:
                    yield v
                # elif str(v.entity).startswith(str(item)):
                #     yield v
                elif str(k).startswith(str(item)):
                    yield v

    def __getitem__(self, item):
        try:
            return next(self._match_key(item))
        except StopIteration:
            raise KeyError('%s' % item)

    '''
    def __getitem__(self, item):
        if isinstance(item, int):
            return
    '''

    def aggregate(self, key=lambda x: x.fragment['StageName'], entity_id=None):
        """
        returns a new LciaResult object in which the components of the original LciaResult object are aggregated into
        static values according to a key.  The key is a lambda expression that is applied to each AggregateLciaScore
        component's entity property (components where the lambda fails will all be grouped together).

        The special key '*' will aggregate all components together.  'entity_id' argument is required in this case to
        provide a distinguishing key for the result (falls back to "aggregated result").

        :param key: default: lambda x: x.fragment['StageName'] -- assuming the payload is a FragmentFlow
        :param entity_id: a descriptive string for the entity, to allow the aggregation to be distinguished in
         subsequent aggregations.  Use 'None' at your peril
        :return:
        """
        agg_result = LciaResult(self.quantity, scenario=self.scenario, private=self._private, scale=self._scale)
        if key == '*':
            if entity_id is None:
                entity_id = 'aggregated result'
            agg_result.add_summary(entity_id, entity_id, 1.0, self.total())
        else:
            for v in self._LciaScores.values():
                keystring = 'other'
                try:
                    keystring = key(v.entity)
                finally:
                    # use keystring AS entity
                    agg_result.add_summary(keystring, keystring, 1.0, v.cumulative_result)
        return agg_result

    def show_agg(self, **kwargs):
        self.aggregate(**kwargs).show_components()  # deliberately don't return anything- or should return grouped?

    def flatten(self, _apply_scale=1.0):
        """
        Return a new LciaResult in which all groupings have been replaced by a set of AggregatedLciaScores, one
         per elementary flow.
        Performs some inline testing via equality assertions, but this still deserves unit testing
        :param: _apply_scale: [1.0] apply a node weighting to the components
        :return:
        """
        flat = LciaResult(self.quantity, scenario=self.scenario, private=self._private, scale=1.0)
        recurse = []  # store flattened summary scores to handle later
        totals = defaultdict(list)
        for k, c in self._LciaScores.items():
            if isinstance(c, SummaryLciaResult):
                if c.static:
                    flat.add_summary(k, c.entity, c.node_weight * _apply_scale, c.unit_score)
                else:
                    recurse.append(c.flatten())
            else:
                for d in c.details():
                    totals[d.factor, d.exchange.direction, d.exchange.termination].append(d.exchange)

        for r in recurse:
            for k in r.keys():
                c = r[k]
                if isinstance(c, SummaryLciaResult):
                    # guaranteed to be static since r is a flattened LciaResult
                    if not c.static:
                        raise InconsistentSummaries(c)
                    try:
                        flat.add_summary(k, c.entity, c.node_weight * _apply_scale, c.unit_score)
                    except InconsistentScores:
                        print('for key %s' % k)
                        raise
                else:
                    for d in c.details():
                        totals[d.factor, d.exchange.direction, d.exchange.termination].append(d.exchange)

        for k, l in totals.items():
            factor, dirn, term = k
            for x in l:
                name = '; '.join([x.flow.name, str(term or factor.context)])
                flat.add_component(name)
                exch = ExchangeValue(x.process, x.flow, dirn,
                                     value=x.value * _apply_scale,
                                     termination=term)
                flat.add_score(name, exch, factor)

        scaled_total = self.total() * _apply_scale
        if not isclose(scaled_total, flat.total(), rel_tol=1e-10):  # this is in lieu of testing, obviously
            print(' LciaResult: %10.4g' % scaled_total)
            print('Flat result: %10.4g' % flat.total())
            print('Difference: %10.4g @ %10.4g' % (flat.total() - scaled_total, _apply_scale))
            if not isclose(scaled_total, flat.total(), rel_tol=1e-6):
                raise ValueError('Total differs by greater than 1e-6! (applied scaling=%10.4g)' % _apply_scale)
        return flat

    @property
    def is_private(self):
        return self._private

    def total(self):
        return sum([i.cumulative_result for i in self._LciaScores.values()])

    def range(self):
        return sum([abs(i.cumulative_result) for i in self._LciaScores.values()])

    def add_component(self, key, entity=None):
        if entity is None:
            entity = key
        if key not in self._LciaScores.keys():
            self._LciaScores[key] = AggregateLciaScore(self, entity)

    def add_score(self, key, exchange, qrresult):
        if qrresult.query != self.quantity:
            raise InconsistentQuantity('%s\nqrresult.quantity: %s\nself.quantity: %s' % (qrresult,
                                                                                         qrresult.query,
                                                                                         self.quantity))
        if key not in self._LciaScores.keys():
            self.add_component(key)
        self._LciaScores[key].add_detailed_result(exchange, qrresult)

    def add_summary(self, key, entity, node_weight, unit_score):
        if key in self._LciaScores.keys():
            # raise DuplicateResult('Key %s is already present' % key)
            '''
            tgt = self._LciaScores[key]
            if isinstance(unit_score, LciaResult):
                uss = unit_score.total()
            else:
                uss = unit_score
            print('Key %s [%s] (%10.4g x %10.4g) adding %s (%10.4g x %10.4g)' % (key,
                                                                                 tgt.entity,
                                                                                 tgt.node_weight, tgt.unit_score,
                                                                                 entity,
                                                                                 node_weight, uss))
            '''
            try:
                self._LciaScores[key] += SummaryLciaResult(self, entity, node_weight, unit_score)
            except InconsistentSummaries:
                self._failed.append(SummaryLciaResult(self, entity, node_weight, unit_score))
        else:
            self._LciaScores[key] = SummaryLciaResult(self, entity, node_weight, unit_score)

    @property
    def failed_summaries(self):
        """
        A list of Summary results that failed to be added to an existing summary. This is mainly diagnostic and should
        be removed soon.
        Note the difiference from self.errors(), which is meant to store input exchanges that could not be converted
        to the query quantity during LCIA.
        :return:
        """
        return self._failed

    def add_cutoff(self, exchange):
        self._cutoffs.append(exchange)

    def cutoffs(self):
        """
        Generates exchanges for which no factor was found during LCIA.
        :return:
        """
        for x in self._cutoffs:
            yield x

    def add_error(self, x, qr):
        self._errors.append(DetailedLciaResult(self, x, qr))

    def errors(self):
        """
        generates exchanges that could not be converted to the target quantity due to a conversion error.
        Note the difference from self.failed_summaries, which reports summary scores that could not be added.
        :return:
        """
        for x in self._errors:
            yield x

    def add_zero(self, x):
        self._zeros.append(x)

    def zeros(self):
        for x in self._zeros:
            yield x

    def details(self):
        for c in self.components():
            for d in c.details():
                yield d

    def keys(self):
        if self._private:
            for k in ():
                yield k
        else:
            for k in self._LciaScores.keys():
                yield k

    def components(self):
        if not self._private:
            for v in self._LciaScores.values():
                yield v

    def component_entities(self):
        if self._private:
            return [None]
        return [k.entity for k in self._LciaScores.values()]

    def _header(self):
        print('%s %s' % (self.quantity, self.unit))
        if self._autorange:
            self.set_autorange()  # update AutoRange object
            print('Auto-ranging: x %g' % self.autorange)
        print('-' * 60)
        if self._scale != 1.0:
            print('%10.4gx %s' % (self._scale, 'scale'))

    def show(self):
        self._header()
        print('%s' % self)

    def show_components(self):
        self._header()
        if not self._private:
            for v in sorted(self._LciaScores.values(), key=lambda x: x.cumulative_result, reverse=True):
                print('%s' % v)
            print('==========')
        print('%s' % self)

    def show_details(self, key=None, **kwargs):
        """
        Sorting by parts is not ideal but it will have to do.
        :param key:
        :param kwargs:
        :return:
        """
        self._header()
        if not self._private:
            if key is None:
                for e in sorted(self._LciaScores.keys(),
                                key=lambda x: self._LciaScores[x].cumulative_result,
                                reverse=True):
                    try:
                        print('\n%s:' % self._LciaScores[e].entity)
                    except TypeError:
                        print('\n%s:' % str(self._LciaScores[e].entity))
                    self._LciaScores[e].show_detailed_result(**kwargs)
            else:
                self._LciaScores[key].show_detailed_result(**kwargs)
        print('%s' % self)

    def __add__(self, other):
        if self.quantity != other.quantity:
            raise InconsistentQuantity
        if self.scenario != other.scenario:
            raise InconsistentScenario
        s = LciaResult(self.quantity, self.scenario)
        for k, v in self._LciaScores.items():
            s._LciaScores[k] = v
            v.update_parent(s)
        for k, v in other._LciaScores.items():
            if k in s._LciaScores:
                if v.entity is s._LciaScores[k].entity:
                    s._LciaScores[k] += v  # this is not implemented yet
                else:
                    s._LciaScores['_%s' % k] = v
            else:
                s._LciaScores[k] = v
        return s

    def __str__(self):
        return '%s %s' % (number(self.total()), self.quantity)

    # charts
    def contrib_query(self, stages=None):
        """
        returns a list of scores
        :param stages: [None] a list of stages to query, or None to return all components.
         Specify '*' to return a 1-item list containing just the total.

        :return:
        """
        if stages == '*':
            return [self.total()]
        elif stages is None:
            stages = self.component_entities()

        data = []
        for c in stages:
            try:
                data.append(self._LciaScores[c].cumulative_result)
            except KeyError:
                data.append(0)
        if not isclose(sum(data), self.total(), rel_tol=1e-6):
            print('Contributions do not equal total [ratio: %.10f]' % (sum(data) / self.total()))
        return data

    def contrib_new(self, *args, autorange=None):
        """
        re-implement contrib query with a better spec.

        Queries are specified as entries from self.keys(). One way to get the keys to be more legible is to first
        perform an aggregation using self.aggregate().

        The current __getitem__ method, which uses a fuzzy match (self._match_keys()) is not currently used.

        :param args: A sequential list of components to query.  The special component '*' can be used to select the
        balance of results.
        :param autorange: [None] do not alter autorange settings.  [True / False]: activate or deactivate auto-ranging.
        :return: a 2-tuple: results, balance where results is a list having the same length as the number of arguments,
         and balance is a float reporting the remainder.  sum(results, balance) == self.total().  If '*' is specified as
         one of the queries, balance will always be 0.
        """
        if autorange is not None:
            self.set_autorange(autorange)
        elif self._autorange is not None:
                self.set_autorange()

        bal_idx = None
        results = []
        for i, query in enumerate(args):
            if query == '*':
                bal_idx = i  # save for later
                results.append(0.0)
            else:
                try:
                    results.append(self._LciaScores[query].cumulative_result * self.autorange)
                except KeyError:
                    results.append(0.0)

        balance = self.total() * self.autorange - sum(results)

        if bal_idx is not None:
            results[bal_idx] = balance
            return results, 0.0
        else:
            return results, balance

    def serialize_components(self, detailed=False):
        return [c.serialize(detailed=detailed) for c in self.components()]



import unittest

from antelope import EntityNotFound

from ...from_json import from_json
from ...archives import Qdb, REF_QTYS
from ...entities.tests.base_testclass import BasicEntityTest

mass_uuid = '93a60a56-a3c8-11da-a746-0800200b9a66'


ref_qs = from_json(REF_QTYS)
num_ref_quantities = len(ref_qs['quantities'])


class QdbTestCase(BasicEntityTest):
    """
    lots to test here
    """
    @classmethod
    def setUpClass(cls):
        super(QdbTestCase, cls).setUpClass()
        cls._qdb = Qdb.new()

    def test_0_qdb_initialization(self):
        self.assertEqual(self._qdb.count_by_type('quantity'), num_ref_quantities)
        self.assertGreaterEqual(num_ref_quantities, 26)  # current number as of 2021-10-30

    def test_mass(self):
        self.assertEqual(self._qdb.query.get_canonical('mass').uuid, mass_uuid)

    def test_canonical(self):
        m = self._qdb.query.get_canonical('mass')
        self.assertEqual(m['Name'], 'Mass')

    def test_kwh_mj_conversion(self):
        elec = next(self._qdb.search('flow', Name='electricity'))
        ncv = self._qdb.query.get_canonical('net calorific value')
        self.assertEqual(ncv.cf(elec), 3.6)

    def test_make_basic_interface(self):
        bi = self._qdb.query.get('93a60a56-a3c8-11da-a746-0800200c9a66')
        self.assertEqual(bi.entity_type, 'quantity')
        self.assertEqual(bi.origin, 'elcd.3.2')

    def test_external_quantity_lookup(self):
        """
        Some quantities should lookup, others should not
        :return:
        """
        for k in self.A.query.quantities():
            if k.external_ref in ('f6811440-ee37-11de-8a39-0800200c9a66', 'e288b5d2-9fcc-4a10-b13c-440786090f43'):
                # 'energy' (unspecified) and 'emissive coolness' respectively
                with self.assertRaises(EntityNotFound):
                    self._qdb.query.get_canonical(k.external_ref)
                self._qdb.add_entity_and_children(k)  # we commented out as foolhardy the code that auto-adds quantities in get_canonical
                self.assertIs(self._qdb.query.get_canonical(k.external_ref), k)
            else:
                qc = self._qdb.query.get_canonical(k)
                self.assertEqual(qc.origin, 'local.qdb')


if __name__ == '__main__':
    unittest.main()



import os
import re
from collections import defaultdict

from .basic_archive import BasicArchive, BASIC_ENTITY_TYPES, InterfaceError, ArchiveError
from .lc_archive import LcArchive, LC_ENTITY_TYPES
from ..implementations import BasicImplementation, IndexImplementation


class AbstractIndex(object):
    def make_interface(self, iface):
        if iface == 'index':
            return IndexImplementation(self)
        elif iface == 'basic':
            return BasicImplementation(self)
        else:
            raise InterfaceError('Unable to create interface %s' % iface)


class BasicIndex(AbstractIndex, BasicArchive):
    _ns_uuid_required = None
    def serialize(self, characterizations=False, values=False, domesticate=False):
        return super(BasicIndex, self).serialize(characterizations=False, values=False, domesticate=False)


class LcIndex(AbstractIndex, LcArchive):
    _ns_uuid_required = None
    def serialize(self, exchanges=False, characterizations=False, values=False, domesticate=False):
        return super(LcIndex, self).serialize(exchanges=False, characterizations=False, values=False, domesticate=False)


def index_archive(archive, source, ref=None, signifier='index', force=False):
    if source is None:
        raise AssertionError('Source is required')
    if not bool(re.search('\.gz$', source)):
        source += '.gz'
    if source == archive.source:
        raise ArchiveError('Index must have a different source from original archive')
    if os.path.exists(source):
        if force:
            print('File exists: %s. Overwriting..' % source)
        else:
            raise FileExistsError(source)

    if ref is None or ref == archive.ref:
        ref = archive.construct_new_ref(signifier=signifier)

    if isinstance(archive, LcArchive):
        index = LcIndex(source, ref=ref, static=True, **archive.init_args)
        types = LC_ENTITY_TYPES
    else:
        index = BasicIndex(source, ref=ref, static=True, **archive.init_args)
        types = BASIC_ENTITY_TYPES
    for t in types:
        for e in archive.entities_by_type(t):
            index.add(e)

    # import original archives list of names
    names = defaultdict(list)
    for s, r in archive.names.items():
        names[r].append(s)

    index.load_from_dict({'catalogNames': names})

    index.write_to_file(source, gzip=True)
    return index



from ...implementations import IndexImplementation
from .q_info import q_info_21 as q_info


class Traci21IndexImplementation(IndexImplementation):
    """
    The main purpose of this is to enable iteration over elements without requiring load_all()
    No need to override lcia methods or quantities- since those all get initialized with the constructor--
    only flowables needs to be init from scratch-- but not right now
    """
    def contexts(self, **kwargs):
        for v in q_info.values():
            self._archive.tm.add_context(v.Compartment, origin=self.origin)
        return super(Traci21IndexImplementation, self).contexts(**kwargs)



"""
Deals with ILCD namespaces

"""

import re
from collections import defaultdict


class XmlWidgetError(Exception):
    pass


def find_ns(nsmap, dtype):
    return next((k for k, v in nsmap.items() if re.search(dtype + '$', v)))


def find_tags(o, tag, ns=None):
    """
    :param o: objectified element
    :param tag:
    :return:
    """
    for found in o.iterfind('.//{0}{1}'.format('{' + o.nsmap[ns] + '}', tag)):
        yield found


def find_tag(o, tag, ns=None):
    """
    :param o: objectified element
    :param tag:
    :return:
    """
    try:
        return next(find_tags(o, tag, ns=ns))
    except StopIteration:
        return ''


def find_common(o, tag):
    return find_tag(o, tag, ns='common')


def _convert_variable_ref(el):
    el = re.sub('%', '%%', str(el))
    return re.sub('{{([A-Za-z0-9_]+)}}', '%(\\1)s', el)


def render_text_block(el, ns=None):
    """
    Created for EcoSpold2.  Take all the children of an objectified element and render them as a text block. Implement
    variable substitution (!!!)
    :param el: Element with tags including 'text' and 'variable'
    :param ns: [None] namespace
    :return:
    """
    vs = list(find_tags(el, 'variable', ns=ns))
    variables = defaultdict(str)
    if len(vs) > 0:
        try:
            for a in vs:
                variables[a.attrib['name']] = a.text
            # variables = {a.attrib['name']: a.text for a in vs}
        except AttributeError:
            print(vs)
            raise
    ts = find_tags(el, 'text', ns=ns)
    texts = [_convert_variable_ref(t) % variables
             for t in sorted(ts, key=lambda _a: _a.attrib['index'])]
    return '\n'.join(texts)



from antelope import comp_dir
from .contexts import Context, NullContext


class ExchangeError(Exception):
    pass


class NoAllocation(Exception):
    pass


class DirectionlessExchangeError(Exception):
    pass


class DuplicateExchangeError(Exception):
    pass


class AmbiguousReferenceError(Exception):
    pass


class Exchange(object):
    """
    An exchange is an affiliation of a process, a flow, and a direction. An exchange does
    not include an exchange value- though presumably a valued exchange would be a subclass.

    An exchange may specify a uuid of a terminating process; tests for equality will distinguish
    differently-terminated flows. (ecoinvent)
    """

    entity_type = 'exchange'

    def __init__(self, process, flow, direction, termination=None, comment=None):
        """

        :param process:
        :param flow:
        :param direction:
        :param termination: external id of terminating process or None (note: this is uuid for ecospold2)
        :param comment: documentary information about the exchange
        :return:
        """
        # assert process.entity_type == 'process', "- we'll allow null exchanges and fragment-terminated exchanges"
        # assert flow.entity_type == 'flow', "'flow' must be an LcFlow"
        # assert direction in directions, "direction must be a string in (%s)" % ', '.join(directions)

        self._process = process
        self._flow = flow
        self._direction = direction
        self._termination = None
        self._comment = comment  # don't bother to serialize these yet...
        if termination is not None:
            if isinstance(termination, str):
                self._termination = termination
            elif isinstance(termination, Context):
                if termination is NullContext:
                    pass
                else:
                    self._termination = termination
                    if termination.origin is None:
                        termination.add_origin(process.origin)

            elif hasattr(termination, 'external_ref'):
                self._termination = termination.external_ref
            else:
                raise ValueError('Unintelligible termination: %s' % termination)
        # self._hash_tuple =
        self._hash = hash((process.external_ref, flow.external_ref, direction, self._termination))
        self._is_reference = False

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, value):
        """
        Comment is cumulative.  This should be generalized to the other entities.. after ContextRefactor...
        :param value:
        :return:
        """
        if self._comment is None:
            self._comment = str(value)
        else:
            self._comment += '\n%s' % str(value)

    @property
    def origin(self):
        return self._process.origin

    @property
    def is_entity(self):
        return self.process.is_entity

    @property
    def is_reference(self):
        return self._is_reference

    def set_ref(self, setter):
        if setter is self._process and (self._termination is None or self.type == 'context'):
            self._is_reference = True
            return True
        return False

    def unset_ref(self, setter):
        if setter is self._process:
            self._is_reference = False
            return True
        return False

    def trim(self):
        return self

    @property
    def unit(self):
        try:
            unit = self.flow.unit
        except AttributeError:
            unit = None
        return unit

    @property
    def value(self):
        return None

    @value.setter
    def value(self, exch_val):
        raise ExchangeError('Cannot set Exchange value')

    """
    These all need to be immutable because they form the exchange's hash
    """
    @property
    def process(self):
        return self._process

    @property
    def flow(self):
        return self._flow

    @property
    def direction(self):
        return self._direction

    @property
    def termination(self):
        return self._termination

    @property
    def key(self):
        return self._hash

    @property
    def lkey(self):
        """
        Long key, for testing equality-- more robust than a hash
        :return:
        """
        return self.flow.external_ref, self._direction, self._termination  # self._hash_tuple

    def is_allocated(self, reference):
        """
        Stub for compatibility
        :param reference:
        :return:
        """
        return False

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if other is None:
            return False
        '''
        if not hasattr(other, 'entity_type'):
            return False
        if other.entity_type != 'exchange':
            return False
        # if self.key == other.key and self.lkey != other.lkey:
        #     raise DuplicateExchangeError('Hash collision!')
        return self.key == other.key
        '''
        try:
            return self.key == other.key
        except AttributeError:
            return False

    def __repr__(self):
        return '%s(%s, %s, %s)' % (self.__class__.__name__, self._process.external_ref,
                                   self.direction, self._tflow)

    @property
    def comp_dir(self):
        return comp_dir(self.direction)

    @property
    def type(self):
        if self.is_reference:
            return 'reference'
        elif self.termination is not None:
            if isinstance(self.termination, Context):
                if self.termination.elementary:
                    return 'elementary'
                return 'context'
            elif self.termination == self.process.external_ref:
                return 'self'
            else:
                return 'node'
        return 'cutoff'

    @property
    def term_ref(self):
        if self.termination is None:
            return None
        if isinstance(self.termination, Context):
            return self.termination.name
        return self.termination

    @property
    def _tflow(self):
        """
        indicates how an exchange is terminated:
         '{*}' - reference
         '   ' - cutoff
         '(=)' - elementary context
         '(-)' - other context
         '(o)' - terminated to self
         '(#)' - terminated to other node
        :return:
        """
        tmark = {
            'reference': '{*} ',
            'cutoff': '    ',
            'self': '(o) ',
            'elementary': '(=) ',
            'context': '(-) ',
            'node': '(#) '
        }[self.type]
        return tmark + str(self.flow)

    def __str__(self):
        return '%s has %s: %s %s' % (self.process, self.direction, self._tflow, self.unit)

    def f_view(self):
        return '%s of %s' % (self.direction, self.process)

    def get_external_ref(self):
        return '%s: %s' % (self.direction, self.flow.external_ref)

    def serialize(self, **kwargs):
        j = {
            'entityType': self.entity_type,
            'flow': self.flow.external_ref,
            'direction': self.direction,
        }
        if self.termination is not None:
            j['termination'] = self.term_ref
        if self.is_reference:
            j['isReference'] = True
        return j

    @classmethod
    def signature_fields(cls):
        return ['process', 'flow', 'direction', 'termination']

    @property
    def args(self):
        return {'comment': self.comment}


class ExchangeValue(Exchange):
    """
    An ExchangeValue is an exchange with a single value (corresponding to unallocated exchange value) plus a dict of
    values allocated to different reference flows.
    """
    '''
    @classmethod
    def from_exchange(cls, exch, value=None, **kwargs):
        if isinstance(exch, ExchangeValue):
            if value is not None:
                raise DuplicateExchangeError('Exchange exists and has value %g (new value %g)' % (exch.value, value))
            return exch
        return cls(exch.process, exch.flow, exch.direction, value=value, **kwargs)
    '''

    @classmethod
    def from_allocated(cls, allocated, reference):
        """
        Use to flatten an allocated process inventory into a standalone inventory
        :param allocated:
        :param reference: a reference exchange
        :return:
        """
        if isinstance(allocated, ExchangeValue):
            return cls(allocated.process, allocated.flow, allocated.direction, value=allocated[reference],
                       termination=allocated.termination, comment=allocated.comment)
        elif isinstance(allocated, Exchange):
            return allocated

    @classmethod
    def from_scenario(cls, allocated, scenario, fallback):
        try:
            value = allocated[scenario]
        except KeyError:
            value = allocated[fallback]
        return cls(allocated.process, allocated.flow, allocated.direction, value=value,
                   termination=allocated.termination)

    def reterminate(self, term=None):
        new = type(self)(self.process, self.flow, self.direction, termination=term, value=self.value,
                         value_dict=self._value_dict)
        if self.comment is not None:
            new.comment = self.comment
        return new

    def add_to_value(self, value, reference=None):
        if reference is None:
            self._value += value
        else:
            if reference in self._value_dict:
                self._value_dict[reference] += value

    def trim(self):
        x = Exchange(self.process, self.flow, self.direction, termination=self.termination)
        if self.is_reference:
            x.set_ref(self.process)
        return x

    def __init__(self, *args, value=None, value_dict=None, **kwargs):
        super(ExchangeValue, self).__init__(*args, **kwargs)
        # assert isinstance(value, float), 'ExchangeValues must be floats (or subclasses)'
        self._value = value
        if value_dict is None:
            self._value_dict = dict()  # keys must live in self.process.reference_entity
        else:
            self._value_dict = value_dict

    @property
    def value(self):
        """
        unallocated value
        :return:
        """
        if self._value is None and len(self._value_dict) == 1:
            return next(v for v in self._value_dict.values())
        return self._value

    @property
    def value_string(self):
        if self._value is None:
            return ' --- '
        return '%.3g' % self._value

    @value.setter
    def value(self, exch_val):
        """
        May only be set once. Otherwise use add_to_value
        :param exch_val:
        :return:
        """
        if self._value is not None:
            raise DuplicateExchangeError('Unallocated exchange value already set to %g (new: %g)' % (self._value,
                                                                                                     exch_val))
        self._value = exch_val

    def is_allocated(self, key):
        """
        Report whether the exchange is allocated with respect to a given reference.
        :param key: an exchange
        :return:
        """
        if len(self._value_dict) > 0:
            return key in self._value_dict
        return False

    def __getitem__(self, item):
        """
        Implements the exchange relation-- computes the quantity of self that is exchanged for a unit of item,
        according to whatever allocation is specified for the exchange or process.  self and item must share the
        same process (tested as external_ref to allow entities and references to interoperate)

        If self is a reference exchange, the exchange relation will equal either 0.0 or 1.0 depending on whether item
        is self or a different reference.  FOR NOW: Reference exchanges cannot

        When an exchange is asked for its value with respect to a particular reference, lookup the allocation in
        the value_dict.  IF there is no value_dict, then the default _value is returned AS LONG AS the process has
        only 0 or 1 reference exchange.

        Allocated exchange values should add up to unallocated value.  When using the exchange values, don't forget to
        normalize by the chosen reference flow's input value (i.e. utilize an inbound exchange when computing
        node weight or when constructing A + B matrices)
        :param item:
        :return:
        """
        if isinstance(item, str) and item in ('process', 'flow', 'termination', 'value'):
            return getattr(self, item)  # %*(#%)(* Marshmallow!
        if not self.process.is_entity:
            return self.process.exchange_relation(item, self.flow, self.direction, self.termination)
        '''
        if len(self._value_dict) == 0:
            # unallocated exchanges always read the same
            return self._value
        '''
        try:
            item = self.process.get_exchange(item.key)
        except KeyError:
            raise ExchangeError('Reference exchange belongs to a different process')
        # elif len(self.process.reference_entity) == 1:
        #    # no allocation necessary
        #    return self.value
        if item.is_reference:
            if self.is_reference:  # if self is a reference entity, the allocation is either .value or 0
                if item.lkey == self.lkey:  # and item.direction == self.direction:
                    return 1.0
                return 0.0
            elif self.process.alloc_qty is not None:
                exch_norm = self.value / self.process.alloc_total  # exchange value per total output quantity
                ref_norm = self.process.alloc_qty.cf(item.flow)  # quantity for a unit output of ref
                return exch_norm * ref_norm
            elif len(self._value_dict) > 0:
                try:
                    return self._value_dict[item] / item.value  # quotient added to correct ecoinvent LCI calcs
                except KeyError:
                    return 0.0
                    # no need to raise on zero allocation
                    # raise NoAllocation('No allocation found for key %s in process %s' % (item, self.process))
                # else fall through
        elif item.termination is not None:
            if not isinstance(item.termination, Context):
                raise ExchangeError('Cannot compute exchange values with respect to a terminated exchange')
        return self._value / item.value

    def __setitem__(self, key, value):
        """
        Used to set custom or "causal" allocation factors.  Here the allocation should be specified as a portion of the
        unallocated exchange value to be allocated to the reference.  The entered values will be normalized by the
        reference exchange's reference value when they are retrieved, to report the exchange value per unit of
        reference.

        This approach makes consistency of causal allocation easy to check: the allocation factors should simply add up
        to the unallocated factor. (this does not apply in ecoinvent, where the different co-products are pre-normalized
        and basically co-inhabiting the same process, and the unallocated factors are not present. The normalization
        must still occur, though, because ecoinvent still does use the *sign* of reference exchanges to indicate
        directionality)
        :param key: a reference exchange
        :param value:
        :return:
        """
        if key.process.external_ref != self.process.external_ref:
            raise ExchangeError('Reference exchange belongs to a different process')
        if not key.is_reference:
            raise AmbiguousReferenceError('Allocation key is not a reference exchange')
        if key in self._value_dict:
            # print(self._value_dict)
            raise DuplicateExchangeError('Exchange value already defined for this reference!')
        if self.is_reference:
            # if it's a reference exchange, it's non-allocatable and should have an empty value_dict
            if self.lkey == key.lkey:
                self.value = value
                # if value != 1:
                #    raise ValueError('Reference Allocation for reference exchange should be 1.0')
            else:
                if value != 0:
                    raise ValueError('Non-reference Allocation for reference exchange should be 0.')
        else:
            self._value_dict[key] = value

    def remove_allocation(self, key):
        """
        Removes allocation if it exists; doesn't complain if it doesn't.
        :param key: a reference exchange.
        :return:
        """
        if key in self._value_dict:
            self._value_dict.pop(key)

    def __str__(self):
        if self.process.entity_type == 'fragment':
            if self.flow == self.process.flow and self.comp_dir == self.process.direction:
                ref = '{*}'
            else:
                ref = '   '
        else:
            if self.is_reference:
                ref = '{*}'
            else:
                ref = '   '
        return '%6.6s: %s [%s %s] %s' % (self.direction, ref, self.value_string, self.unit, self._tflow)

    def f_view(self):
        return '%6.6s: [%s %s] %s' % (self.direction, self.value_string, self.unit, self.process)

    def _serialize_value_dict(self):
        j = dict()
        for k, v in self._value_dict.items():
            j['%s:%s' % (k.direction, k.flow.external_ref)] = v
        return j

    def serialize(self, values=False):
        j = super(ExchangeValue, self).serialize()
        if values:
            if self.value is not None:
                j['value'] = float(self.value)
            if not self.is_reference and len(self._value_dict) > 0:
                j['valueDict'] = self._serialize_value_dict()
        return j


class DissipationExchange(ExchangeValue):
    """
    Composition / Dissipation mechanics can be encapsulated entirely into the exchange object- the problem
     is reduced to a serialization / deserialization problem.

    Composition / Dissipation probably conflicts with Allocation, i.e. it is not supported to add a dissipation
    factor to an AllocatedExchange.

    This is a problem because the allocated exchange is also the secret sauce being used to implement processflow
    parameters- and would presumably be the same mechanism to implement dissipation parameters.

    best solution would be to fold reference spec higher up in inheritance- also need an ironclad way to determine
     the input flow (or to not support dissipation for processes with more than one reference flow)
    and simply make flow_quantity not None be the key to interpret the exchange value as a dissipation rate.
    except that I don't want to squash the non-dissipation exchange value.

    Maybe I should not be worrying about this right now.

    """
    def __init__(self, *args, flow_quantity=None, scale=1.0, dissipation=1.0, value=None, **kwargs):
        self.flow_quantity = flow_quantity
        self.scale = scale
        self.dissipation = dissipation
        super(ExchangeValue, self).__init__(*args, **kwargs)
        self._value = value  # used only when dissipation is not defined

    def content(self, ref_flow=None):
        """
        :param ref_flow: a flow LcEntity
        :return:
        """
        if ref_flow is None:
            ref_flow = list(self.process.reference_entity)[0]
            if len(self.process.reference_entity) > 1:
                raise AmbiguousReferenceError
        if ref_flow.cf(self.flow_quantity) != 0:
            return ref_flow.cf(self.flow_quantity)
        return None

    @property
    def value(self):
        c = self.content()
        if c is not None:
            return c * self.scale * self.dissipation
        return self._value

    def __str__(self):
        raise NotImplemented

    def serialize(self, values=False):
        raise NotImplemented


class MarketExchange(Exchange):
    """
    A MarketExchange is an alternative implementation of an ExchangeValue that handles the multiple-input process
    case, i.e. when several processes produce the same product, and the database must balance them according to
    some apportionment of market value or production volume.

    The client code has to explicitly create a market exchange.  How does it know to do that? in the case of
    ecospold2, it has to determine whether the process has duplicate [non-zero] flows with activityLinkIds.

    In other cases, it will be foreground / linker code that does it.

    Add market suppliers using dictionary notation.  Use exchange values or production volumes, but do it consistently.
    The exchange value returned is always the individual supplier's value divided by the sum of values.
    """
    def __init__(self, *args, **kwargs):
        super(MarketExchange, self).__init__(*args, **kwargs)
        self._market_dict = dict()

    def _sum(self):
        return sum([v for k, v in self._market_dict.items()])

    def keys(self):
        return self._market_dict.keys()

    def markets(self):
        return self._market_dict.items()

    def __setitem__(self, key, value):
        if key in self._market_dict:
            raise KeyError('Key already exists with value %g (new value %g)' % (self._market_dict[key], value))
        self._market_dict[key] = value

    def __getitem__(self, item):
        return self._market_dict[item] / self._sum()

    def __str__(self):
        return 'Market for %s: %d suppliers, %g total volume' % (self.flow, len(self._market_dict), self._sum())

    def serialize(self, values=False):
        j = super(MarketExchange, self).serialize()
        if values:
            j['marketSuppliers'] = self._market_dict
        else:
            j['marketSuppliers'] = [k for k in self.keys()]
        return j



from ..lc_resource import LcResource
from collections import namedtuple

ResourceInfo = namedtuple('ResourceInfo',
                          ('source', 'ds_type', 'download_url', 'download_md5sum', 'config', 'init_args'))


class DataSourceError(Exception):
    pass


class DataSource(object):
    """
    An abstract class that defines how data sources are handled.
    """
    _ds_type = None

    def __init__(self, data_root=None, **kwargs):
        self._root = data_root
        self._kwargs = kwargs

    @property
    def root(self):
        return self._root

    def _make_resource(self, ref, source=None, ds_type=None, info=None, **kwargs):
        if isinstance(info, ResourceInfo):
            source = info.source
            ds_type = info.ds_type
            if info.download_url is not None:
                kwargs['download'] = {'url': info.download_url, 'md5sum': info.download_md5sum}
            if info.config is not None:
                kwargs['config'] = info.config
            if info.init_args is not None:
                kwargs.update(info.init_args)
        if ds_type is None:
            ds_type = self._ds_type
        return LcResource(ref, source, ds_type, **kwargs)

    def register_all_resources(self, cat):
        for ref in self.references:
            for res in self.make_resources(ref):
                cat.add_resource(res)

    @property
    def references(self):
        """
        Generates a list of semantic references the DataSource knows how to instantiate
        :return:
        """
        raise NotImplementedError

    def interfaces(self, ref):
        """
        Generates a list of interfaces known for the given reference. the reference must be in the list of references.
        :param ref:
        :return:
        """
        raise NotImplementedError

    def make_resources(self, ref):
        """
        Generates an exhaustive sequence of LcResource objects for a given reference.
        :param ref:
        :return:
        """
        raise NotImplementedError

    def test_params(self, ref, interface):
        """
        User supplies a reference and interface.
        This function returns a 2-tuple:
          - kwargs to pass as input arguments
          - expected output

        This function can't be written until the tests are designed.
        :param ref:
        :param interface:
        :return:
        """
        fun = {
            'index': self._index_test_params,
            'inven': self._inventory_test_params,
            'backg': self._background_test_params,
            'quant': self._quantity_test_params
        }[interface[:5]]
        return fun(ref)

    def _index_test_params(self, ref):
        pass

    def _inventory_test_params(self, ref):
        pass

    def _background_test_params(self, ref):
        pass

    def _quantity_test_params(self, ref):
        pass


class DataCollection(DataSource):
    """
    A container for a number of related data sources.  Uses a factory to generate the data sources, and then spools
    out their results.
    """
    def factory(self, *args, **kwargs):
        raise NotImplementedError

    def __init__(self, data_root, **kwargs):
        super(DataCollection, self).__init__(data_root, **kwargs)
        self._sources = dict()
        for b in self.factory(data_root, **kwargs):
            for r in b.references:
                if r in self._sources:
                    raise KeyError('Duplicate reference %s' % r)
                self._sources[r] = b

    @property
    def references(self):
        for s in self._sources.keys():
            yield s

    def interfaces(self, ref):
        b = self._sources[ref]
        for i in b.interfaces(ref):
            yield i

    def make_resources(self, ref):
        b = self._sources[ref]
        for m in b.make_resources(ref):
            yield m



import unittest

from ..lcia_results import LciaResult
from antelope import CatalogRef


class LciaResultTestCase(unittest.TestCase):
    def test_catalog_ref(self):
        """
        Create an LCIA result using a catalog ref as quantity
        :return:
        """
        qty = CatalogRef('fictitious.origin', 'dummy_ext_ref', entity_type='quantity', Name='Test Quantity')
        res = LciaResult(qty)
        self.assertEqual(res.total(), 0.0)


if __name__ == '__main__':
    unittest.main()



import ast

def parse_math(expression):
    """
    A function I got off stackoverflow that enables python to parse user input as a mathematical expression.
    probably a huge security risk. but it enables the user to enter missing characterization values during runtime.
    :param expression:
    :return:
    """
    try:
        tree = ast.parse(expression, mode='eval')
    except SyntaxError:
        return    # not a Python expression
    if not all(isinstance(node, (ast.Expression, ast.UnaryOp, ast.unaryop, ast.BinOp, ast.operator, ast.Num))
               for node in ast.walk(tree)):
        return    # not a mathematical expression (numbers and operators)
    return eval(compile(tree, filename='', mode='eval'))



from .data_source import DataSource
import os


EI_LCIA_SPREADSHEETS = {
    '3.1': 'LCIA implementation v3.1 2014_08_13.xlsx',
    '3.4': 'LCIA_implementation_3.4.xlsx',
    '3.5': 'LCIA_implementation_3.5.xlsx',
    '3.6': 'LCIA_implementation_3.6.xlsx',
    '3.7': 'LCIA_implementation_3.7.1.xlsx',
    '3.8': 'LCIA Implementation v3.8.xlsx',
}


E_CFG = {'hints': [  # cover elementary contexts that need directional hints
    ['context', 'air', 'to air'],
    ['context', 'water', 'to water']
]}


class EcoinventLciaConfig(DataSource):
    _ifaces = ('index', 'quantity')
    _ds_type = 'EcoinventLcia'

    def __init__(self, data_root, version, **kwargs):
        self._version = str(version)
        try:
            self._sourcefile = EI_LCIA_SPREADSHEETS[self._version]
        except KeyError:
            self._sourcefile = None
        super(EcoinventLciaConfig, self).__init__(data_root, **kwargs)

    @property
    def references(self):
        if self._sourcefile is not None:
            if os.path.exists(os.path.join(self._root, self._sourcefile)):
                yield '.'.join(['local', 'lcia', 'ecoinvent', self._version])

    def interfaces(self, ref):
        for k in self._ifaces:
            yield k

    def make_resources(self, ref):
        if ref not in self.references:
            raise ValueError('Unknown reference %s' % ref)
        if self._sourcefile is None:
            raise AttributeError('This exception should never occur')  # because self.references screens self._info
        source = os.path.join(self._root, self._sourcefile)
        yield self._make_resource(ref, source=source, interfaces=self._ifaces, version=self._version, static=True,
                                  config=E_CFG)



"""
Code for generating a catalog populated with local data sources.  This is where active workflow development can
occur.  See GitHub root / workflow.txt for thoughts.

MAIN IDEA:
 local.py:
  The local config file. specifies source locations and resource-specific configurations for data sources.
  It must define the following variables:
   CATALOG_ROOT = location to build a permanent data catalog
   TEST_ROOT = location to store temporary test materials (blown away after every use)
   RESOURCES_CONFIG = a dict mapping source nicknames to instantiation parameters.
    The instantiation parameters have two required keys:
     'source': key whose value is an executable DataSource subclass
     'enable_test': key whose absence (or falsity) will suppress tests.
    All other kv pairs in the dict are passed as input arguments to the DataSource.

 data_source.py:
  provides an abstract class for specifying data sources.  Each data source should be defined as a
  subclass of DataSource that implements its interface.

 tests/test_local.py:
  uses local.py to generate and validate the persistent catalog at the location specified in CATALOG_ROOT.

TO ADD A NEW DATA SOURCE:

 1. Produce a resource factory for your data source.

 1a. Create or locate an appropriate DataSource subclass, and implement its abstract methods:
     references: generate a list of semantic references the class knows how to instantiate
     interfaces(ref): generate a list of interfaces known for the given ref (ref must be in references)
     make_resources(ref): generate an exhaustive sequence of LcResource objects for a given ref

 1b. Make sure the resources created by your DataSource subclass include any necessary configuration information

 2. Provide configuration info for your data source

 2a. Import your DataSource subclass here.

 2b. Create an entry in RESOURCES_CONFIG with at minimum the 'source' and 'enable_test' keys, as well as any
     necessary input arguments.

 3. Run python -m unittest lcatools/data_sources/tests/test_local.py

"""

from .gwp_ipcc_2007 import GwpIpcc2007
from .traci import TraciConfig
from .uslci import UsLciConfig

# local resources
#from .ecoinvent import EcoinventConfig
#from .ecoinvent_lcia import EcoinventLciaConfig
#from .calrecycle_lca import CalRecycleConfig


'''RESOURCES_CONFIG 
provides enabling and configuration information to the various data resource
classes defined elsewhere in this package.  Each key should match a python module name, and each value should
be a dict.
 
The contents of the module dict are resource-specific but must include a 'source' field which maps to a config object,
and a 'data_root' which maps to the _root variable in the DataSource.
The remaining args are passed as init args to the object.

The RESOURCES_CONFIG dict by default includes all the resources the catalog can make without any site-specific
configuration.  Use RESOURCES_CONFIG_LOCAL to add and configure local resources.  Still figuring out the best way 
to actually do all this.  Community input welcome. 
'''

RESOURCES_CONFIG = {
    'ipcc2007': {
        'source': GwpIpcc2007,
        'enable_test': True
    },
    'uslci': {
        'source': UsLciConfig,
        'enable_test': True
    },
    'traci': {
        'source': TraciConfig,
        'data_root': '/data/LCI/TRACI/',
        'enable_test': True
    }
}


"""
RESOURCES_CONFIG_LOCAL = {
    'ecoinvent': {
        'source': EcoinventConfig,
        'data_root': '/data/LCI/Ecoinvent/',
        'enable_test': False
    },
    'calrecycle': {
        'source': CalRecycleConfig,
        'data_root': '/data/GitHub/CalRecycle/LCA_Data/',
        'enable_test': False
    },
    'ecoinvent_lcia': {
        'source': EcoinventLciaConfig,
        'version': '3.1',
        'data_root': '/data/LCI/Ecoinvent/LCIA/',
        'enable_test': False
    }
}

RESOURCES_CONFIG.update(RESOURCES_CONFIG_LOCAL)  # still trying to figure out the best way to do this
"""

'''OPERATIONAL CONFIG
Code below this line is used by the local init machinery to setup catalogs / testing
'''


def make_config(resource=None):
    if resource is None:
        print('Configurable Resources:')
        for k, v in RESOURCES_CONFIG.items():
            print('%s: %s' % (k, v['source'].__name__))
        return None
    d = RESOURCES_CONFIG[resource]
    return d['source'](**{k: v for k, v in d.items() if k != 'source'})


def check_enabled(resource):
    if resource in RESOURCES_CONFIG:
        if 'enable_test' in RESOURCES_CONFIG[resource]:
            return RESOURCES_CONFIG[resource]['enable_test']
    return False



from ...implementations import IndexImplementation


class EcoSpold2IndexImplementation(IndexImplementation):
    def re_index(self, cutoffs=False):
        """
        do nothing here as termination index is not subject to change.
        :param cutoffs:
        :return:
        """
        pass

    def targets(self, flow, **kwargs):
        for p in self._terminations[flow]:
            yield self._archive.retrieve_or_fetch_entity(p)

    def processes(self, literal=False, **kwargs):
        for p in self._archive.processes(**kwargs):
            yield p



from __future__ import print_function, unicode_literals

import os
from itertools import chain

from lxml import objectify
from lxml.etree import XMLSyntaxError

from urllib.error import HTTPError


from ..file_store import FileStore
from ..xml_widgets import *

from ...archives import LcArchive, uuid_regex
from ...entities import LcFlow, LcProcess, LcQuantity, LcUnit
from ...characterizations import DuplicateCharacterizationError

typeDirs = {'Process': 'processes',
            'Flow': 'flows',
            'LCIAMethod': 'LCIAmethods',  # love that case consistency
            'FlowProperty': 'flowproperties',
            'UnitGroup': 'unitgroups',
            'Source': 'sources',
            'Contact': 'contacts'
            }

elcd3_local_fallback = os.path.join(os.path.expanduser('~'), 'Dropbox', 'data',
                                    'ELCD', 'ELCD3.2.zip')


elcd3_remote_fallback = "http://eplca.jrc.ec.europa.eu/ELCD3/resource/"


def _check_dtype(dtype):
    if dtype not in typeDirs:
        print('Datatype %s not known.' % dtype)
        return False
    return True


def _extract_uuid(filename):
    return uuid_regex.search(filename).groups()[0]


def _extract_dtype(filename, pathtype=os.path):
    cands = [i for i in re.split(pathtype.sep, filename) if i in typeDirs.values()]
    dtype = [k for k, v in typeDirs.items() if v in cands]
    if len(dtype) == 0:
        raise ValueError('No dtype found: %s' % filename)
    return dtype[0]


def dtype_from_nsmap(nsmap):
    for v in nsmap.values():
        cand = re.sub('(^.*/)', '', v)
        if cand in typeDirs:
            return cand
    return None


def get_flow_ref(exch, ns=None):
    f_uuid = find_tag(exch, 'referenceToFlowDataSet', ns=ns).attrib['refObjectId']
    f_uri = find_tag(exch, 'referenceToFlowDataSet', ns=ns).attrib['uri']
    f_dir = find_tag(exch, 'exchangeDirection', ns=ns).text
    return f_uuid, f_uri, f_dir


def grab_flow_name(o, ns=None):
    return ', '.join(chain(filter(len, [str(find_tag(o, k, ns=ns))
                                        for k in ('baseName',
                                                  'treatmentStandardsRoutes',
                                                  'mixAndLocationTypes',
                                                  'flowProperties')])))


def get_reference_flow(process, ns=None):
    try_ref = find_tag(process, 'referenceToReferenceFlow', ns=ns)
    if try_ref == '':
        return None, None, None  # multioutput, no specified reference
    else:
        ref_to_ref = int(try_ref)
    rf = [i for i in process['exchanges'].getchildren()
          if int(i.attrib['dataSetInternalID']) == ref_to_ref][0]
    return get_flow_ref(rf, ns=ns)


def get_reference_flow_property_id(flow, ns=None):
    # load or check the reference quantity
    return int(find_tag(flow, 'referenceToReferenceFlowProperty', ns=ns))
'''    rfp = [i for i in flow['flowProperties'].getchildren()
           if int(i.attrib['dataSetInternalID']) == ref_to_ref][0]
    rfp_uuid = find_tag(rfp, 'referenceToFlowPropertyDataSet', ns=ns)[0].attrib['refObjectId']
    rfp_uri = find_tag(rfp, 'referenceToFlowPropertyDataSet', ns=ns)[0].attrib['uri']
    return rfp_uuid, rfp_uri
'''


def get_reference_unit_group(q, ns=None):
    ref_to_ref = find_tag(q, 'referenceToReferenceUnitGroup', ns=ns)
    ug_uuid = ref_to_ref.attrib['refObjectId']
    ug_uri = ref_to_ref.attrib['uri']
    return ug_uuid, ug_uri


def get_exch_value(exch, ns=None):
    try:
        v = float(find_tag(exch, 'resultingAmount', ns=ns))
    except ValueError:
        v = None
    return v

'''
class IlcdEntity(object):
    """
    Container for objectified ILCD entities from XML
    """
    def __init__(self, path):
        """

        :param path:
        :return:
        """
        self.o = objectify.fromstring(string)

    def el(self):
'''


class IlcdArchive(LcArchive):
    """
    This class handles de-referencing for ILCD archives
    """

    def __init__(self, source, prefix=None, **kwargs):
        """
        Just instantiates the parent class.
        :param source: root of the archive
        :param prefix: difference between the internal path (ref) and the ILCD base
          (note: for local archives, this defaults to 'ILCD'; for remote arcnives it
           defaults to empty)
        :param quiet: forwarded to ArchiveInterface
        :return:
        """
        super(IlcdArchive, self).__init__(source, **kwargs)
        self.internal_prefix = prefix
        if prefix is not None:
            self._serialize_dict['prefix'] = prefix

        self._archive = FileStore(self.source, internal_prefix=prefix)

        if not self._archive.OK:
            print('Trying local ELCD reference')
            self._archive = FileStore(elcd3_local_fallback)
            if self._archive.OK:
                self._source = elcd3_local_fallback
        if not self._archive.OK:
            print('Falling back to ELCD Remote Reference')
            self._archive = FileStore(elcd3_remote_fallback, query_string='format=xml')
            if self._archive.OK:
                self._source = elcd3_remote_fallback
        if not self._archive.remote:
            self._archive.internal_prefix = 'ILCD'  # appends

    @property
    def _pathtype(self):
        return self._archive.pathtype

    @staticmethod
    def _path_from_ref(ref):
        """
        This fails if the filename has a version specification
        :param ref:
        :return:
        """
        return ref + '.xml'

    @staticmethod
    def _path_from_uri(uri):
        """
        just need to strip any leading '../' from the uri
        :param uri:
        :return:
        """
        uri = re.sub('^(\.\./)*', '', uri)
        return uri

    def _path_from_parts(self, dtype, uid, version=None):
        """
        aka 'path from parts'
        :param dtype: required
        :param uid: required
        :param version: optional [None]
        :return: a single (prefixed) path
        """
        assert _check_dtype(dtype)
        postpath = self._pathtype.join(typeDirs[dtype], uid)
        if version is not None:
            postpath += '_' + version
        return postpath + '.xml'

    def search_by_id(self, uid, dtype=None):
        return [i for i in self.list_objects(dtype=dtype) if re.search(uid, i, flags=re.IGNORECASE)]

    def list_objects(self, dtype=None):
        assert self._archive.remote is False, "Cannot list objects for remote archives"
        in_prefix = None
        if dtype is not None:
            in_prefix = typeDirs[dtype]
        for f in self._archive.listfiles(in_prefix=in_prefix):
            yield f

    def _fetch_filename(self, filename):
        return self._archive.readfile(filename)

    def _check_or_retrieve_child(self, uid, uri):
        child = self.__getitem__(uid)  # looks upstream first
        if child is None:
            child = self._fetch(uid, uri=uri)
        return child

    def _get_objectified_entity(self, filename):
        return objectify.fromstring(self._fetch_filename(filename))

    def _search_for_term(self, term, dtype=None):
        search_results = self.search_by_id(term, dtype=dtype)
        if len(search_results) > 0:
            self._print('Found %d Results:' % len(search_results))
            if len(search_results) < 100:
                [print(i) for i in search_results]
            if len(search_results) > 1:
                print('Please refine search')
                return None
            result = search_results[0]
            dtype = _extract_dtype(result, self._pathtype)
            return self.objectify(result, dtype=dtype)
        print('No results for %s' % term)
        return None

    def objectify(self, term, dtype=None, version=None, uri=None):
        if uri is not None:
            return self._get_objectified_entity(self._path_from_uri(uri))
        if dtype is None:
            try:
                dtype = _extract_dtype(term, self._pathtype)
            except ValueError:
                return self._search_for_term(term)

        try:
            uid = _extract_uuid(term)
        except AttributeError:
            # can't find UUID: search is required
            return self._search_for_term(term, dtype=dtype)

        # if we get here, uid is valid and dtype is valid
        entity = self.__getitem__(uid)  # checks upstream first (!! should this be local only?)
        if entity is not None:
            return entity

        try:
            # if we are a search result, this will succeed
            o = self._get_objectified_entity(term)
        except (KeyError, FileNotFoundError):
            # we are not a search result-- let's build the entity path
            try:
                o = self._get_objectified_entity(self._path_from_parts(dtype, uid, version=version))
            except (KeyError, FileNotFoundError):
                # still not found- maybe it isn't here
                return None
        return o

    def _create_unit(self, unit_ref):
        """
        UnitGroups aren't stored as full-fledged entities- they are stored as dicts inside quantities.
        :param unit_ref:
        :return:
        """
        dtype = _extract_dtype(unit_ref, self._pathtype)
        try:
            uid = _extract_uuid(unit_ref)
        except AttributeError:
            return super(IlcdArchive, self)._create_unit(unit_ref)
        filename = self._path_from_parts(dtype, uid)
        o = self._get_objectified_entity(filename)

        ns = find_ns(o.nsmap, 'UnitGroup')

        u = str(find_common(o, 'UUID'))
        reference_unit = int(find_tag(o, 'referenceToReferenceUnit', ns=ns))
        unitstring = str(o['units'].getchildren()[reference_unit]['name'])
        ref_unit = LcUnit(unitstring, unit_uuid=u)
        ref_unit.set_external_ref('%s/%s' % (typeDirs['UnitGroup'], u))

        unitconv = dict()
        for i in o['units'].getchildren():
            unitconv[str(i['name'])] = 1.0 / float(i['meanValue'])
        return ref_unit, unitconv

    def _create_quantity(self, o):
        """

        :param o: objectified FlowProperty
        :return:
        """
        u = str(find_common(o, 'UUID'))
        try_q = self[u]
        if try_q is not None:
            return try_q

        ns = find_ns(o.nsmap, 'FlowProperty')

        n = str(find_common(o, 'name'))

        c = str(find_common(o, 'generalComment'))

        ug, ug_uri = get_reference_unit_group(o, ns=ns)

        ug_path = self._pathtype.join('unitgroups', ug)  # need the path without extension- I know- it's all sloppy

        refunit, unitconv = self._create_unit(ug_path)

        ext_ref = '%s/%s' % (typeDirs['FlowProperty'], u)

        q = LcQuantity(ext_ref, Name=n, ReferenceUnit=refunit, UnitConversion=unitconv, Comment=c, uuid=u)

        self.add(q)

        return q

    @staticmethod
    def _create_dummy_flow_from_exch(uid, exch):
        n = str(find_common(exch, 'shortDescription'))
        print('Creating DUMMY flow (%s) with name %s' % (uid, n))
        return LcFlow(uid, Name=n, Comment='Dummy flow (HTTP or XML error)', Compartment=['dummy flows'])

    def _create_flow(self, o):
        """

        :param o: objectified flow
        :return: an LcFlow
        """
        u = str(find_common(o, 'UUID'))
        try_f = self[u]
        if try_f is not None:
            return try_f

        ns = find_ns(o.nsmap, 'Flow')
        n = grab_flow_name(o, ns=ns)

        c = str(find_common(o, 'generalComment'))

        cas = str(find_tag(o, 'CASNumber', ns=ns))

        cat = [str(i) for i in find_tags(o, 'category', ns='common')]
        if len(cat) == 0:
            cat = [str(i) for i in find_tags(o, 'class', ns='common')]

        ref_to_ref = get_reference_flow_property_id(o, ns=ns)
        chars = []
        ref_qty = None

        for fp in o['flowProperties'].getchildren():

            ref = find_tag(fp, 'referenceToFlowPropertyDataSet', ns=ns)
            rfp_uuid = ref.attrib['refObjectId']
            rfp_uri = ref.attrib['uri']

            try:
                q = self._check_or_retrieve_child(rfp_uuid, rfp_uri)
            except (HTTPError, XMLSyntaxError, FileNotFoundError):  # ??
                print('%s: skipping characterization for %s' % (u, rfp_uuid))
                continue

            if int(fp.attrib['dataSetInternalID']) == ref_to_ref:
                ref_qty = q
            else:
                val = float(find_tag(fp, 'meanValue', ns=ns))
                chars.append((q, val))

        ext_ref = '%s/%s' % (typeDirs['Flow'], u)
        if str(find_tag(o, 'typeOfDataSet', ns=ns)) == 'Elementary flow':
            f = LcFlow(ext_ref, Name=n, CasNumber=cas, Comment=c, ReferenceQuantity=ref_qty, Compartment=cat)
        else:
            f = LcFlow(ext_ref, Name=n, CasNumber=cas, Comment=c, ReferenceQuantity=ref_qty, Class=cat)

        self.add(f)

        for qty, val in chars:
            try:
                self.tm.add_characterization(f.link, f.reference_entity, qty, val, context=f.context)
            except DuplicateCharacterizationError:
                print('Duplicate Characterization in entity %s\n %s = %g' % (u, qty, val))
                # let it go

        return f

    def _create_process_entity(self, o, ns):
        u = str(find_common(o, 'UUID'))
        try_p = self[u]
        if try_p is not None:
            return try_p

        n = ', '.join(chain(filter(len, [str(find_tag(o, k, ns=ns))
                                         for k in ('baseName',
                                                   'treatmentStandardsRoutes',
                                                   'mixAndLocationTypes',
                                                   'functionalUnitFlowProperties')])))

        try:
            g = find_tag(o, 'locationOfOperationSupplyOrProduction', ns=ns).attrib['location']
        except AttributeError:
            g = 'GLO'

        stt = {'begin': str(find_common(o, 'referenceYear')), 'end': str(find_common(o, 'dataSetValidUntil'))}

        c = str(find_common(o, 'generalComment'))

        cls = [str(i) for i in find_tags(o, 'class', ns='common')]

        ext_ref = '%s/%s' % (typeDirs['Process'], u)

        p = LcProcess(ext_ref, Name=n, Comment=c, SpatialScope=g, TemporalScope=stt,
                      Classifications=cls, uuid=u)

        self.add(p)

        return p

    def _create_process(self, o):
        """

        :param o: objectified process
        :return:
        """
        ns = find_ns(o.nsmap, 'Process')

        try:
            rf, rf_uri, rf_dir = get_reference_flow(o, ns=ns)
        except XMLSyntaxError:
            rf = None
            rf_dir = None

        exch_list = []

        for exch in o['exchanges'].getchildren():
            # load all child flows
            f_id, f_uri, f_dir = get_flow_ref(exch, ns=ns)
            try:
                f = self._check_or_retrieve_child(f_id, f_uri)
            except (HTTPError, XMLSyntaxError, FileNotFoundError):
                u = str(find_common(o, 'UUID'))
                print('In UUID %s:' % u)
                f = self._create_dummy_flow_from_exch(f_id, exch)
                self.add(f)
            v = get_exch_value(exch, ns=ns)
            cmt = find_tag(exch, 'generalComment', ns=ns)
            exch_list.append((f, f_dir, v, cmt))

        p = self._create_process_entity(o, ns)

        for flow, f_dir, val, cmt in exch_list:
            cx = self.tm[flow.context]
            x = p.add_exchange(flow, f_dir, reference=None, value=val, termination=cx,  # no intermediate terms in ILCD
                               add_dups=True)  # add_dups: poor quality control on ELCD
            if len(cmt) > 0:
                x.comment = cmt
            if rf == flow.uuid and rf_dir == f_dir:
                p.set_reference(flow, f_dir)

        return p

    def _fetch(self, term, dtype=None, version=None, **kwargs):
        """
        fetch an object from the archive by reference.

        term is either: a uid and a dtype (and optional version) OR a filename
        dtype MUST be specified as kwarg for remote archives; otherwise will search
        :param term:
        :return:
        """
        if dtype is None:
            try:
                dtype = _extract_dtype(term, self._pathtype)
            except ValueError:
                pass

        if term is None or len(term) == 0:
            return None

        o = self.objectify(term, dtype=dtype, version=version, **kwargs)
        if o is None:
            return None

        if dtype is None:
            dtype = dtype_from_nsmap(o.nsmap)

        if dtype == 'Flow':
            try:
                return self._create_flow(o)
            except KeyError:
                print('KeyError on term %s dtype %s version %s' % (term, dtype, version))
        elif dtype == 'Process':
            return self._create_process(o)
        elif dtype == 'FlowProperty':
            return self._create_quantity(o)
        else:
            return o

    def _load_all(self):
        for i in self.list_objects('Process'):
            self.retrieve_or_fetch_entity(i)
        self.check_counter('quantity')
        self.check_counter('flow')
        self.check_counter('process')


'''
class IlcdWebInterface(IlcdArchive):
    """
    Collects entities from an ILCD web server and stores their metadata locally

    self.ref should be assumed to be an http link prefix _ such that:
    _/(datatypes)/(uuid).xml is valid for readable uuids.

    """
    def fetch_entity(self, type, id):
        """

        :param type:
        :param id:
        :return:
        """
'''



from antelope import ExchangeInterface, EntityNotFound
from .basic import BasicImplementation


class ExchangeImplementation(BasicImplementation, ExchangeInterface):
    """
    This provides access to detailed exchange values and computes the exchange relation.
    Creates no additional requirements on the archive.
    """
    def exchanges(self, process, **kwargs):
        p = self._archive.retrieve_or_fetch_entity(process)
        for x in p.exchanges():
            yield x

    def exchange_values(self, process, flow, direction=None, termination=None, reference=None, **kwargs):
        if reference is True:
            for x in self.get_reference(process):
                if x.flow.external_ref == flow:
                    yield x
        else:
            p = self._archive.retrieve_or_fetch_entity(process)
            for x in p.exchange_values(self.get(flow), direction=direction):
                if reference is False and x.is_reference:
                    continue
                if termination is None:
                    yield x
                else:
                    if x.termination == termination:
                        yield x

    def inventory(self, process, ref_flow=None, scenario=None, **kwargs):
        p = self._archive.retrieve_or_fetch_entity(process)
        if p is None:
            raise EntityNotFound(process)
        if p.entity_type == 'process':
            '''
            for x in sorted(p.inventory(ref_flow=ref_flow),
                            key=lambda t: (not t.is_reference, t.direction, t.value or 0.0)):
            '''
            for x in p.inventory(ref_flow=ref_flow, **kwargs):
                yield x
        elif p.entity_type == 'fragment':
            for x in p.inventory(scenario=scenario, observed=True, **kwargs):
                yield x

    def exchange_relation(self, process, ref_flow, exch_flow, direction, termination=None, **kwargs):
        """
        This certainly should be tested for quantity-based and ecoinvent-style (exhaustive) allocation

        :param process:
        :param ref_flow:
        :param exch_flow:
        :param direction:
        :param termination:
        :return:
        """
        p = self._archive.retrieve_or_fetch_entity(process)
        norm = p.reference(ref_flow)
        if termination is None:
            xs = [x for x in p.exchange_values(flow=exch_flow, direction=direction)]
            if len(xs) == 1:
                return xs[0][norm]
            elif len(xs) == 0:
                return 0.0
            else:
                return sum(x[norm] for x in xs)
        else:
            x = p.get_exchange(hash((p.external_ref, exch_flow, direction, termination)))
            return x[norm]



"""
Each archive now has a TermManager which interprets query arguments as synonyms for canonical flows and contexts.  This
can also be upgraded to an LciaEngine, which extends the synonymization strategy to quantities as well
"""
from antelope import (QuantityInterface, NoFactorsFound, ConversionReferenceMismatch, EntityNotFound, FlowInterface,
                      convert, NoUnitConversionTable, QuantityRequired)

from .basic import BasicImplementation
from ..characterizations import QRResult, LocaleMismatch
from ..contexts import NullContext
from ..lcia_results import LciaResult


class RefQuantityRequired(Exception):
    pass


class QuantityConversion(object):
    """
    A stack of Quantity Relation results that are composed sequentially in order to render a flow-quantity conversion.
    The first QRR added should report the query quantity (numerator) in terms of some reference quantity (denominator);
    then each subsequent QRR should include the prior ref quantity as the query quantity.

    The QuantityConversion has a subset of the interface of a QRResult (flowable, ref, query, context, value), leaving
    out locale and origin for the time being since they could vary across factors.

    For instance, a Quantity conversion from moles of CH4 to GWP 100 might include first the GWP conversion and then
    the mol conversion:
    QuantityConversion(QRResult('methane', 'kg', 'kg CO2eq', 'emissions to air', 'GLO', 'ipcc.2007', 25.0),
                       QRResult('methane', 'mol', 'kg', None, 'GLO', 'local.qdb', 0.016))
    giving the resulting value of 0.4.

    The QuantityConversion needs information to be fully defined: the query quantity and the query context, both of
    which should be canonical.  The canonical context is especially needed to test directionality for LCIA.
    """
    @classmethod
    def null(cls, flowable, rq, qq, context, locale, origin):
        qrr = QRResult(flowable, rq, qq, context or NullContext, locale, origin, 0.0)
        return cls(qrr)

    @classmethod
    def copy(cls, conv):
        return cls(*conv.results, query=conv.query, context=conv.context)

    def __init__(self, *args, query=None, context=NullContext):
        self._query = query  # this is just a stub to give ref conversion machinery something to grab hold of
        self._context = context
        self._results = []
        for arg in args:
            self.add_result(arg)

    def __hash__(self):
        return hash((self.flowable, self.ref, self.query, self.context, self.locale))

    def __eq__(self, other):
        try:
            return (self.flowable == other.flowable and self.query == other.query and self.context == other.context
                    and self.locale == other.locale and self.value == other.value)
        except AttributeError:
            return False

    def invert(self):
        inv_qrr = type(self)(query=self.ref)
        for res in self._results[::-1]:
            inv_qrr.add_inverted_result(res)
        return inv_qrr

    def flatten(self, origin=None):
        if len(self._results) == 0:
            return self
        if origin is None:
            origin = self._results[0].origin
        return QRResult(self.flowable, self.ref, self.query, self.context, self.locale, origin, self.value)

    def add_result(self, qrr):
        if isinstance(qrr, QRResult):
            if qrr.query is None or qrr.ref is None:
                raise ValueError('Both ref and query quantity must be defined')
            if len(self._results) > 0:
                if self.flowable != qrr.flowable:  # don't think we care
                    # raise FlowableMismatch('%s != %s' % (self.flowable, qrr.flowable))
                    print('Flowable changed: %s -> %s' % (self.flowable, qrr.flowable))
                if self.ref != qrr.query:
                    raise ConversionReferenceMismatch('%s != %s' % (self.ref, qrr.query))
            self._results.append(qrr)
        else:
            raise TypeError('Must supply a QRResult')

    @property
    def qualitative(self):
        return isinstance(self.value, str)

    @property
    def query(self):
        if self._query is None:
            if len(self._results) > 0:
                return self._results[0].query
        return self._query

    @property
    def ref(self):
        if len(self._results) == 0:
            return self._query
        return self._results[-1].ref

    @property
    def flowable(self):
        if len(self._results) == 0:
            return None
        return self._results[0].flowable

    @property
    def context(self):
        for qrr in reversed(self._results):
            if qrr.context is not None:
                return qrr.context
        if self._context is not None:
            return self._context
        return NullContext

    @property
    def locale(self):
        locs = []
        for res in self._results:
            if res.locale not in locs:
                locs.append(res.locale)
        return '/'.join(locs)

    @property
    def origin(self):
        if len(self._results) > 0:
            return self._results[0].origin
        return ''

    @property
    def results(self):
        for res in self._results:
            yield res

    def seen(self, q):
        for res in self._results:
            if q == res.ref:
                return True
        return False

    @staticmethod
    def _invert_qrr(qrr):
        """
        swaps the ref and query quantities and inverts the value
        :param qrr:
        :return:
        """
        return QRResult(qrr.flowable, qrr.query, qrr.ref, qrr.context, qrr.locale, qrr.origin, 1.0 / qrr.value)

    def add_inverted_result(self, qrri):
        self.add_result(self._invert_qrr(qrri))

    @property
    def value(self):
        val = 1.0
        for res in self._results:
            if isinstance(res.value, str):
                return res.value  # qualitative value
            val *= res.value
        return val

    def __getitem__(self, item):
        return self._results[item]

    def __str__(self):
        if self.qualitative:
            return '%s [%s] %s: %s [%s] (%s)' % (self.flowable, self.context, self.query, self.value,
                                                 self._results[-1].locale, self._results[-1].origin)
        if len(self._results) == 0:
            return str(self.query)
        conv = ' x '.join(['%g %s/%s' % (res.value, res.query.unit, res.ref.unit) for res in self._results])
        return '%s; %s: %s [%s] (%s)' % (self.flowable, self.context,
                                         conv, self._results[-1].locale,
                                         self._results[-1].origin)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__str__())

    # TODO: add serialization, other outputs


class QuantityConversionError(object):
    def __init__(self, qrresult, ref_quantity):
        self._qrr = qrresult
        self._ref = ref_quantity

    @property
    def ref(self):
        return self._ref

    @property
    def query(self):
        return self._qrr.query

    @property
    def fail(self):
        return self._qrr.ref

    @property
    def flowable(self):
        return self._qrr.flowable

    @property
    def context(self):
        return self._qrr.context

    @property
    def locale(self):
        return self._qrr.locale

    @property
    def value(self):
        return None

    def __repr__(self):
        return '%s(%s; %s %s =X=> %s)' % (self.__class__.__name__, self.flowable, self.context, self._qrr.ref, self._ref)


class NoConversion(Exception):
    pass


def try_convert(flowable, rq, qq, context, locale):

    if hasattr(qq, 'is_lcia_method') and qq.is_lcia_method:
        raise NoConversion
    else:
        try:
            fac = convert(qq, from_unit=rq.unit)
            return QRResult(flowable, rq, qq, context or NullContext, locale, qq.origin, fac)
        except (KeyError, NoUnitConversionTable):
            pass

    if hasattr(rq, 'is_lcia_method') and rq.is_lcia_method:
        raise NoConversion
    else:
        try:
            fac = convert(rq, to=qq.unit)
            return QRResult(flowable, rq, qq, context or NullContext, locale, rq.origin, fac)
        except (KeyError, NoUnitConversionTable):
            pass
    raise NoConversion


class QuantityImplementation(BasicImplementation, QuantityInterface):
    """
    Uses the archive's term manager to index cfs, by way of the canonical quantities
    """
    def quantities(self, **kwargs):
        for q_e in self._archive.search('quantity', **kwargs):
            yield q_e

    def get_canonical(self, quantity, **kwargs):
        """
        Retrieve a canonical quantity from a qdb; else raises EntityNotFound
        :param quantity: external_id of quantity
        :return: quantity entity
        """
        try:
            return self._archive.tm.get_canonical(quantity)
        except EntityNotFound:
            if isinstance(quantity, str):
                q = self._archive.retrieve_or_fetch_entity(quantity)
                if q is None:
                    raise
                return self._archive.tm.get_canonical(q.external_ref)
#            elif hasattr(quantity, 'entity_type') and quantity.entity_type == 'quantity':
#                self._archive.add_entity_and_children(quantity)
#                return self._archive.tm.get_canonical(quantity)
            else:
                raise

    def factors(self, quantity, flowable=None, context=None, dist=0):
        q = self.get_canonical(quantity)
        for cf in self._archive.tm.factors_for_quantity(q, flowable=flowable, context=context, dist=dist):
            yield cf

    def characterize(self, flowable, ref_quantity, query_quantity, value, context=None, location='GLO', origin=None,
                     **kwargs):
        """
        We gotta be able to do this
        :param flowable: string
        :param ref_quantity: string
        :param query_quantity: string
        :param value: float
        :param context: string
        :param location: string
        :param origin: [self.origin]
        :param kwargs: overwrite=False
        :return:
        """
        rq = self.get_canonical(ref_quantity)
        qq = self.get_canonical(query_quantity)
        if qq.origin != self._archive.ref:
            self._archive.add_entity_and_children(qq)  # catches EntityExists
        if origin is None:
            origin = self.origin
        return self._archive.tm.add_characterization(flowable, rq, qq, value, context=context, location=location,
                                                     origin=origin, **kwargs)

    def _ref_qty_conversion(self, target_quantity, flowable, compartment, conv, locale, _reverse=True):
        """
        Transforms a CF into a quantity conversion with the proper ref quantity. Does it recursively! watch with terror.
        :param target_quantity: conversion target
        :param flowable:
        :param compartment:
        :param conv: An existing QuantityConversion chain, whose ref we must turn into target_quantity
        :param locale:
        :return: the incoming conv, augmented
        """
        if target_quantity is None:
            raise ConversionReferenceMismatch('Cannot convert to None')
        found_quantity = self.get_canonical(conv.ref)  # this is still necessary because CFs are stored with native ref quantities
        if found_quantity != target_quantity:
            # zero look for conversions
            try:
                qr_result = try_convert(flowable, target_quantity, found_quantity, compartment, locale)
                conv.add_result(qr_result)
                return conv
            except NoConversion:
                pass

            # first look for forward matches
            cfs_fwd = [cf for cf in self._archive.tm.factors_for_flowable(flowable, quantity=found_quantity,
                                                                          context=compartment, dist=3)
                       if not conv.seen(cf.ref_quantity)]
            for cf in cfs_fwd:
                new_conv = QuantityConversion.copy(conv)
                new_conv.add_result(cf.query(locale))
                try:
                    return self._ref_qty_conversion(target_quantity, flowable, compartment, new_conv, locale,
                                                    _reverse=_reverse)
                except ConversionReferenceMismatch:
                    continue

            # then look for reverse matches... but... only once
            if _reverse:
                new_conv = QuantityConversion.copy(conv)
                try:
                    rev_conv = self._ref_qty_conversion(found_quantity, flowable, compartment,
                                                        QuantityConversion(query=target_quantity), locale,
                                                        _reverse=False)

                    for res in rev_conv.invert().results:
                        new_conv.add_result(res)
                    if new_conv.ref == target_quantity:
                        return new_conv
                except QuantityRequired:  # if we don't have a reverse qty interface, we can't look for reverse cfs
                    pass

            raise ConversionReferenceMismatch('Flow %s\nfrom %s\nto %s' % (flowable,
                                                                           conv.ref,
                                                                           target_quantity))
        return conv

    def _quantity_engine(self, fb, rq, qq, cx, locale='GLO',
                         **kwargs):
        """
        This is the main "comprehensive" engine for performing characterizations.

        :param fb: a string that is synonymous with a known flowable.
        :param rq: a canonical ref_quantity or None
        :param qq: a canonical query_quantity or None
        :param cx: a known context or None
        :param locale: ['GLO']
        :param kwargs:
         dist: CLookup distance (0=exact 1=subcompartments 2=parent 3=all parents)
        :return: 3-tuple of lists of QRResults objects: qr_results, qr_mismatch, qr_geog
         qr_results: valid conversions from query quantity to ref quantity
         qr_geog: valid conversions which had a broader spatial scope than specified, if at least one narrow result
         qr_mismatch: conversions from query quantity to a different quantity that could not be further converted

        """
        # TODO: port qdb functionality: detect unity conversions; quell biogenic co2; integrate convert() -- DONE??

        qr_results = []
        qr_mismatch = []
        qr_geog = []

        # if we are not doing LCIA, jump straight to unit conversion
        if qq is not None:
            if not qq.is_lcia_method:
                res = QuantityConversion(query=qq, context=cx)
                try:
                    qr_results.append(self._ref_qty_conversion(rq, fb, cx, res, locale))
                except ConversionReferenceMismatch:
                    res.add_result(QRResult(fb, qq, qq, cx, locale, qq.origin, 1.0))
                    qr_mismatch.append(QuantityConversionError(res, rq))
                except LocaleMismatch as e:
                    locales = e.args[0]
                    for loc in locales:
                        res = QuantityConversion(query=qq, context=cx)
                        qr_geog.append(self._ref_qty_conversion(rq, fb, cx, res, loc))

                return qr_results, qr_geog, qr_mismatch

        for cf in self._archive.tm.factors_for_flowable(fb, quantity=qq, context=cx, **kwargs):
            try:
                res = QuantityConversion(cf.query(locale), query=qq, context=cx)
                qr_results.append(self._ref_qty_conversion(rq, fb, cx, res, locale))
            except ConversionReferenceMismatch:
                qr_mismatch.append(QuantityConversionError(res, rq))
            except LocaleMismatch as e:
                locales = e.args[0]
                for loc in locales:
                    qr_geog.append(self._ref_qty_conversion(rq, fb, cx, res, loc))

        ''' # leaving this OUT- we should only do forward and reverse matching for ref quantity conversion, not qq
        Leaving it back in because it breaks a unit test
        and backing it out again because of minor _ref_qty_conversion refactor
        for cf in self._archive.tm.factors_for_flowable(fb, quantity=rq, context=cx, **kwargs):
            res = QuantityConversion(cf.query(locale))
            try:
                qr_results.append(self._ref_qty_conversion(qq, fb, cx, res, locale).invert())
            except ConversionReferenceMismatch:
                pass  # qr_mismatch.append(res.invert())  We shouldn't be surprised that there is no reverse conversion
        ##'''

        if len(qr_results + qr_geog + qr_mismatch) == 0:
            raise NoFactorsFound

        if len(qr_results) > 1:
            _geog = [k for k in filter(lambda x: x[0].locale != locale, qr_results)]
            qr_results = [k for k in filter(lambda x: x[0].locale == locale, qr_results)]
            qr_geog += _geog

        return qr_results, qr_geog, qr_mismatch

    def _get_flowable_info(self, flow, ref_quantity, context):
        """
        We need all three defined at the end. So if all given, we take em and look em up.
        Basically we take what we get for flow, unless we're missing ref_quantity or context.
        If flow is not entity_type='flow', we try to fetch a flow and if that fails we return what we were given.
        If we're given a flow and/or an external ref that looks up, then flowable, ref qty, and context are taken from
        it (unless they were provided)
        If a context query is specified and no canonical context is found, NullContext is returned (matches only
        NullContext characterizations).
        Otherwise, None is returned (matches everything).
        :param flow:
        :param ref_quantity:
        :param context:
        :return: flowable, canonical ref_quantity, canonical context or None
        """
        if isinstance(flow, str):
            try:
                flow = self.get(flow)  # assume it's a ref
            except EntityNotFound:
                pass

            # if f is not None:  # lookup succeeded
            #     flow = f

        if isinstance(flow, FlowInterface):
            flowable = flow.link
            if ref_quantity is None:
                ref_quantity = flow.reference_entity
                if ref_quantity is None:
                    print('a FlowInterface? %s\n%s' % (flow, flow.__class__.__name__))
                    print('No ref quantity? %s' % flow.link)

            if context is None:
                context = flow.context or None
        else:
            if hasattr(flow, 'link'):
                flowable = flow.link
            else:
                flowable = str(flow)

        if ref_quantity is None:
            raise RefQuantityRequired

        rq = self.get_canonical(ref_quantity)
        cx = self._archive.tm[context]  # will fall back to find_matching_context if tm is an LciaEngine
        if cx is None and context is not None:  # lcia_engine now returns NullContext, but TermManager does not
            cx = NullContext
        return flowable, rq, cx

    def quantity_conversions(self, flow, query_quantity, ref_quantity=None, context=None, locale='GLO', **kwargs):
        """
        Return a comprehensive set of conversion results for the provided inputs.  This method catches errors and
        returns a null result if no factors are found.

        This function is a wrapper to handle inputs.

        :param flow: a string that is synonymous with a flowable characterized by the query quantity
        :param query_quantity: convert to this quantity
        :param ref_quantity: [None] convert for 1 unit of this quantity
        :param context: [None] a string synonym for a context / "archetype"? (<== locale-specific?)
        :param locale: handled by CF; default 'GLO'
        :param kwargs:
         dist: CLookup distance (0=exact 1=subcompartments 2=parent compartment 3=all parents)
        :return: a 3-tuple of lists of QuantityConversion objects:
         [valid conversions],
         [geographic proxy conversions],
         [mismatched ref unit conversions]
        """
        flowable, rq, cx = self._get_flowable_info(flow, ref_quantity, context)
        if query_quantity is None:
            qq = None
        else:
            qq = self.get_canonical(query_quantity)
            if qq is None:
                raise EntityNotFound(qq)

        if qq == rq:  # is?
            return [QRResult(flowable, rq, qq, context or NullContext, locale, qq.origin, 1.0)], [], []

        try:
            return self._quantity_engine(flowable, rq, qq, cx, locale=locale, **kwargs)
        except NoFactorsFound:
            if qq is None:
                return [], [], []
            else:
                return [QuantityConversion.null(flowable, rq, qq, cx, locale, self.origin)], [], []

    def _quantity_relation(self, fb, rq, qq, cx, locale='GLO',
                           strategy=None, allow_proxy=True, **kwargs):
        """
        Reports the first / best result of a quantity conversion.  Returns a single QRResult interface
        (QuantityConversion result) that converts unit of the reference quantity into the query quantity for the given
        fb, context, and locale (default 'GLO').
        If the locale is not found, this would be a great place to run a spatial best-match algorithm.

        :param fb:
        :param rq:
        :param qq:
        :param cx:
        :param locale:
        :param strategy: approach for resolving multiple-CF in dist>0.  ('highest' | 'lowest' | 'average' | ...? )
          None = return first result
        :param allow_proxy: [True] in the event of 0 exact results but >0 geographic proxies, return a geographic
          proxy without error.
        :param kwargs:
        :return: a QRResult object or interface
        """
        try:
            qr_results, qr_geog, qr_mismatch = self._quantity_engine(fb, rq, qq, cx,
                                                                     locale=locale, **kwargs)
        except NoFactorsFound:
            qr_results, qr_geog, qr_mismatch = [QuantityConversion.null(fb, rq, qq, cx, locale, self.origin)], [], []

        if len(qr_results) == 0 and len(qr_geog) > 0:
            if allow_proxy:
                qr_results += qr_geog
            else:
                locs = ', '.join([qr.locale for qr in qr_geog])
                print('allow_proxy to show values with locales: %s' % locs)

        if len(qr_results) > 1:
            # this is obviously punting
            if strategy == 'first':
                best = qr_results[0]
            elif strategy is None or strategy == 'highest':  # highest CF is most conservative
                val = max(v.value for v in qr_results)
                best = next(v for v in qr_results if v.value == val)
            elif strategy == 'lowest':
                val = min(v.value for v in qr_results)
                best = next(v for v in qr_results if v.value == val)
            # elif strategy == 'average':
            #     return sum(v.value for v in qr_results) / len(qr_results)
            else:
                raise ValueError('Unknown strategy %s' % strategy)
        elif len(qr_results) == 1:
            best = qr_results[0]
        else:
            if len(qr_mismatch) == 0:
                raise NoFactorsFound

            best = None
        return best, qr_mismatch

    def quantity_relation(self, flowable, ref_quantity, query_quantity, context, locale='GLO',
                          strategy=None, allow_proxy=True, **kwargs):

        fb, rq, cx = self._get_flowable_info(flowable, ref_quantity, context)  # call only for exception handling
        qq = self.get_canonical(query_quantity)

        if qq == rq:  # is?
            return QRResult(flowable, rq, qq, context or NullContext, locale, qq.origin, 1.0)

        result, mismatch = self._quantity_relation(fb, rq, qq, cx, locale=locale,
                                                   strategy=strategy, allow_proxy=allow_proxy, **kwargs)
        if result is None:
            if len(mismatch) > 0:
                '''
                for k in mismatch:
                    print('Conversion failure: Flowable: %s\nfrom: %s %s\nto: %s %s' % (fb, k.fail, k.fail.link,
                                                                                        rq, rq.link))
                '''

                raise ConversionReferenceMismatch

            else:
                raise AssertionError('Something went wrong')
        return result

    def cf(self, flow, quantity, ref_quantity=None, context=None, locale='GLO', **kwargs):
        """
        Should Always return a number and catch errors
        :param flow:
        :param quantity:
        :param ref_quantity: [None] taken from flow.reference_entity if flow is entity or locally known external_ref
        :param context: [None] taken from flow.reference_entity if flow is entity or locally known external_ref
        :param locale:
        :param kwargs: allow_proxy [False], strategy ['first'] -> passed to quantity_relation
        :return: the value of the QRResult found by the quantity_relation
        """
        try:
            qr = self.quantity_relation(flow, ref_quantity, quantity, context=context, locale=locale, **kwargs)
            return qr.value
        except ConversionReferenceMismatch:
            return 0.0

    def flat_qr(self, flow, quantity, ref_quantity=None, context=None, locale='GLO', **kwargs):
        val = self.quantity_relation(flow, ref_quantity, quantity, context, locale=locale, **kwargs)
        return val.flatten(origin=self.origin)

    def profile(self, flow, ref_quantity=None, context=None, complete=False, **kwargs):
        """
        Generate characterizations for the named flow or flowable.  The positional argument is first used to retrieve
        a flow, and if successful, the reference quantity and context are taken for that flow.  Otherwise, the
        positional argument is interpreted as a flowable synonym and used to generate CFs, optionally filtered by
        context.  In that case, if no ref quantity is given then the CFs are returned as-reported; if a ref quantity is
        given then a ref quantity conversion is attempted and the resulting QRResult objects are returned.

        This whole interface desperately needs testing.

        :param flow:
        :param ref_quantity: [None]
        :param context: [None]
        :param complete: [False] if True, report all results including errors and geographic proxies
        :param kwargs:
        :return:
        """
        qrr, qrg, qrm = self.quantity_conversions(flow, None, ref_quantity, context, **kwargs)

        for r in qrr:
            yield r

        if complete:
            for r in qrg + qrm:
                yield r

    def norm(self, quantity, region=None, **kwargs):
        """

        :param quantity:
        :param region:
        :param kwargs:
        :return:
        """
        q = self.get_canonical(quantity)
        n = q.get('normalisationFactors', [])
        if len(n) == 0:
            return 0.0
        ix = 0
        if region is not None:
            try:
                ix = next(i for i, k in enumerate(q.get('normSets', [])) if k == region)
            except StopIteration:
                pass
        return n[ix]

    def _lookup_x(self, x, q, locale, refresh=False, **kwargs):
        cx = self._archive.tm[x.termination]
        if refresh:  # destroy prior matches, even if no new one is found (only overwrite would not truly "refresh")
            x.flow.pop_char(q, cx, locale)
        else:
            try:  # look for a cached result
                return x.flow.chk_char(q, cx, locale)
            except KeyError:
                pass
        try:
            ref_q = self.get_canonical(x.flow.reference_entity)
        except EntityNotFound:
            ref_q = self._archive.tm.add_quantity(x.flow.reference_entity)
        try:
            qr, mis = self._quantity_relation(x.flow.name, ref_q, q, cx, locale=locale,
                                              **kwargs)
            if qr is None and len(mis) > 0:
                qrr = mis[0]
                if q.is_lcia_method:
                    print('Conversion error %s' % x)
                    if len(mis) > 1:
                        print('omitting mismatches:')
                        for k in mis[1:]:
                            print(k)
            else:
                qrr = qr
        except NoFactorsFound:
            qrr = None

        ## if not x.flow.is_entity:  # why don't we cache conversions on actual flows? because cfs might change??
        x.flow.see_char(q, cx, locale, qrr)
        return qrr

    def do_lcia(self, quantity, inventory, locale='GLO', group=None, dist=2, **kwargs):
        """
        Successively implement the quantity relation over an iterable of exchanges.

        man, WHAT is the qdb DOING with all those LOC? (ans: seemingly a lot)

        :param quantity:
        :param inventory: An iterable of exchange-like entries, having flow, direction, value, termination.  Currently
          also uses process.external_ref for hashing purposes, but that could conceivably be abandoned.
        :param locale: ['GLO']
        :param group: How to group scores.  Should be a lambda that operates on inventory items. Default x -> x.process
        :param dist: [2] controls how strictly to interpret exchange context.
          0 - exact context matches only;
          1 - match child contexts (code default)
          2 - match parent contexts [this default]
          3 - match any ancestor context, including Null
         exchange context; 3 - match any ancestor of exch
        :param kwargs:
        :return:
        """
        q = self.get_canonical(quantity)
        res = LciaResult(q)
        if group is None:
            group = lambda _x: _x.process
        for x in inventory:
            xt = x.type
            if xt in ('cutoff', 'reference'):
                res.add_cutoff(x)
                continue
            elif xt == 'self':
                continue
            qrr = self._lookup_x(x, q, locale, dist=dist, **kwargs)
            if isinstance(qrr, QuantityConversion):
                if qrr.value == 0:
                    res.add_zero(x)
                else:
                    res.add_score(group(x), x, qrr)
            elif isinstance(qrr, QuantityConversionError):
                res.add_error(x, qrr)
            elif qrr is None:
                res.add_cutoff(x)
            else:
                raise TypeError('Unknown qrr type %s' % qrr)
        return res

    def lcia(self, process, ref_flow, quantity_ref, **kwargs):
        """
        Implementation of foreground LCIA -- moved from LcCatalog
        :param process:
        :param ref_flow:
        :param quantity_ref:
        :param kwargs:
        :return:
        """
        p = self._archive.retrieve_or_fetch_entity(process)
        return quantity_ref.do_lcia(p.inventory(ref_flow=ref_flow),
                                    locale=p['SpatialScope'])

    def fragment_lcia(self, fragment, quantity_ref, scenario=None, refresh=False, **kwargs):
        frag = self._archive.retrieve_or_fetch_entity(fragment)
        return frag.top().fragment_lcia(quantity_ref, scenario=scenario, refresh=refresh, **kwargs)

